import ctypes
import time
import warnings

try:
    from memory_profiler import memory_usage
except ImportError:
    memory_usage = None
from pytest import mark, raises

from wand.color import Color
from wand.version import MAGICK_VERSION_INFO, QUANTUM_DEPTH  # noqa


def test_user_error():
    with raises(TypeError):
        Color()
    with raises(ValueError):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            Color('not_a_color')


def test_user_raw():
    with Color('black') as black:
        with Color(raw=black.raw) as c:
            assert c


def test_user_pickle():
    with Color('black') as black:
        # Can't trust string literal between IM versions, but the tuple
        # should be constant.
        assert len(black.__getinitargs__()) == 2


def test_equals():
    """Equality test."""
    assert Color('#fff') == Color('#ffffff') == Color('white')
    assert Color('#000') == Color('#000000') == Color('black')
    assert Color('rgba(0, 0, 0, 0)') == Color('rgba(0, 0, 0, 0)')
    assert Color('rgba(0, 0, 0, 0)') == Color('rgba(1, 1, 1, 0)')
    assert Color('green') != 'green'


def test_not_equals():
    """Equality test."""
    assert Color('#000') != Color('#fff')
    assert Color('rgba(0, 0, 0, 0)') != Color('rgba(0, 0, 0, 1)')
    assert Color('rgba(0, 0, 0, 1)') != Color('rgba(1, 1, 1, 1)')


def test_hash():
    """Hash test."""
    assert (hash(Color('#fff')) == hash(Color('#ffffff')) ==
            hash(Color('white')))
    assert (hash(Color('#000')) == hash(Color('#000000')) ==
            hash(Color('black')))
    assert hash(Color('rgba(0, 0, 0, 0))')) == hash(Color('rgba(0, 0, 0, 0))'))
    assert hash(Color('rgba(0, 0, 0, 0))')) == hash(Color('rgba(1, 1, 1, 0))'))


def test_red():
    assert Color('black').red == 0
    assert Color('red').red == 1
    assert Color('white').red == 1
    assert 0.5 <= Color('rgba(128, 0, 0, 1)').red < 0.51
    c = Color('none')
    c.red = 1
    assert c.red == 1


def test_green():
    assert Color('black').green == 0
    assert Color('#0f0').green == 1
    assert Color('white').green == 1
    assert 0.5 <= Color('rgba(0, 128, 0, 1)').green < 0.51
    c = Color('none')
    c.green = 1
    assert c.green == 1


def test_blue():
    assert Color('black').blue == 0
    assert Color('blue').blue == 1
    assert Color('white').blue == 1
    assert 0.5 <= Color('rgba(0, 0, 128, 1)').blue < 0.51
    c = Color('none')
    c.blue = 1
    assert c.blue == 1


def test_alpha():
    assert Color('rgba(0, 0, 0, 1)').alpha == 1
    assert Color('rgba(0, 0, 0, 0)').alpha == 0
    assert 0.49 <= Color('rgba(0, 0, 0, 0.5)').alpha <= 0.51
    c = Color('none')
    c.alpha = 1
    assert c.alpha == 1


def test_cyan():
    assert Color('cmyk(100%, 0, 0, 0)').cyan == 1
    assert Color('cmyk(0, 0, 0, 0)').cyan == 0
    c = Color('none')
    c.cyan = 1
    assert c.cyan == 1


def test_magenta():
    assert Color('cmyk(0, 100%, 0, 0)').magenta == 1
    assert Color('cmyk(0, 0, 0, 0)').magenta == 0
    c = Color('none')
    c.magenta = 1
    assert c.magenta == 1


def test_yellow():
    assert Color('cmyk(0, 0, 100%, 0)').yellow == 1
    assert Color('cmyk(0, 0, 0, 0)').yellow == 0
    c = Color('none')
    c.yellow = 1
    assert c.yellow == 1


def test_black():
    assert Color('cmyk(0, 0, 0, 100%)').black == 1
    assert Color('cmyk(0, 0, 0, 0)').black == 0
    c = Color('none')
    c.black = 1
    assert c.black == 1


def test_red_quantum():
    q = 2 ** QUANTUM_DEPTH - 1
    assert Color('black').red_quantum == 0
    assert Color('red').red_quantum == q
    assert Color('white').red_quantum == q
    assert (0.49 * q) < Color('rgba(128, 0, 0, 1)').red_quantum < (0.51 * q)
    c = Color('none')
    c.red_quantum = q
    assert c.red_quantum == q


def test_green_quantum():
    q = 2 ** QUANTUM_DEPTH - 1
    assert Color('black').green_quantum == 0
    assert Color('#0f0').green_quantum == q
    assert Color('white').green_quantum == q
    assert (0.49 * q) < Color('rgba(0, 128, 0, 1)').green_quantum < (0.51 * q)
    c = Color('none')
    c.green_quantum = q
    assert c.green_quantum == q


def test_blue_quantum():
    q = 2 ** QUANTUM_DEPTH - 1
    assert Color('black').blue_quantum == 0
    assert Color('blue').blue_quantum == q
    assert Color('white').blue_quantum == q
    assert (0.49 * q) < Color('rgba(0, 0, 128, 1)').blue_quantum < (0.51 * q)
    c = Color('none')
    c.blue_quantum = q
    assert c.blue_quantum == q


def test_alpha_quantum():
    q = 2 ** QUANTUM_DEPTH - 1
    assert Color('rgba(0, 0, 0, 1)').alpha_quantum == q
    assert Color('rgba(0, 0, 0, 0)').alpha_quantum == 0
    assert (0.49 * q) < Color('rgba(0, 0, 0, 0.5)').alpha_quantum < (0.51 * q)
    c = Color('none')
    c.alpha_quantum = q
    assert c.alpha_quantum == q


def test_cyan_quantum():
    q = 2 ** QUANTUM_DEPTH - 1
    assert int(Color('cmyk(100%, 0, 0, 0)').cyan_quantum) == q
    assert Color('cmyk(0, 0, 0, 0)').cyan_quantum == 0
    c = Color('none')
    c.cyan_quantum = q
    assert c.cyan_quantum == q


def test_magenta_quantum():
    q = 2 ** QUANTUM_DEPTH - 1
    assert int(Color('cmyk(0, 100%, 0, 0)').magenta_quantum) == q
    assert Color('cmyk(0, 0, 0, 0)').magenta_quantum == 0
    c = Color('none')
    c.magenta_quantum = q
    assert c.magenta_quantum == q


def test_yellow_quantum():
    q = 2 ** QUANTUM_DEPTH - 1
    assert int(Color('cmyk(0, 0, 100%, 0)').yellow_quantum) == q
    assert Color('cmyk(0, 0, 0, 0)').yellow_quantum == 0
    c = Color('none')
    c.yellow_quantum = q
    assert c.yellow_quantum == q


def test_black_quantum():
    q = 2 ** QUANTUM_DEPTH - 1
    assert int(Color('cmyk(0, 0, 0, 100%)').black_quantum) == q
    assert Color('cmyk(0, 0, 0, 0)').black_quantum == 0
    c = Color('none')
    c.black_quantum = q
    assert c.black_quantum == q


def test_red_int8():
    assert Color('black').red_int8 == 0
    assert Color('red').red_int8 == 255
    assert Color('white').red_int8 == 255
    assert Color('rgba(128, 0, 0, 1)').red_int8 == 128
    c = Color('none')
    c.red_int8 = 255
    assert c.red_int8 == 255


def test_green_int8():
    assert Color('black').green_int8 == 0
    assert Color('#0f0').green_int8 == 255
    assert Color('white').green_int8 == 255
    assert Color('rgba(0, 128, 0, 1)').green_int8 == 128
    c = Color('none')
    c.green_int8 = 255
    assert c.green_int8 == 255


def test_blue_int8():
    assert Color('black').blue_int8 == 0
    assert Color('blue').blue_int8 == 255
    assert Color('white').blue_int8 == 255
    assert Color('rgba(0, 0, 128, 1)').blue_int8 == 128
    c = Color('none')
    c.blue_int8 = 255
    assert c.blue_int8 == 255


def test_alpha_int8():
    assert Color('rgba(0, 0, 0, 1)').alpha_int8 == 255
    assert Color('rgba(0, 0, 0, 0)').alpha_int8 == 0
    if not (Color('rgb(127,0,0)').red_quantum <=
            Color('rgba(0,0,0,0.5').alpha_quantum <=
            Color('rgb(128,0,0)').red_quantum):
        # FIXME: I don't know why, but the value PixelGetAlphaQuantum() returns
        #        is inconsistent to other PixelGet{Red,Green,Blue}Quantum()
        #        functions in Travis CI.  We just skip the test in this case.
        return
    assert 127 <= Color('rgba(0, 0, 0, 0.5)').alpha_int8 <= 128
    c = Color('none')
    c.alpha_int8 = 255
    assert c.alpha_int8 == 255


def test_cyan_int8():
    assert Color('cmyk(100%, 0, 0, 0)').cyan_int8 == 255
    assert Color('cmyk(0, 0, 0, 0)').cyan_int8 == 0
    c = Color('none')
    c.cyan_int8 = 255
    assert c.cyan_int8 == 255


def test_magenta_int8():
    assert Color('cmyk(0, 100%, 0, 0)').magenta_int8 == 255
    assert Color('cmyk(0, 0, 0, 0)').magenta_int8 == 0
    c = Color('none')
    c.magenta_int8 = 255
    assert c.magenta_int8 == 255


def test_yellow_int8():
    assert Color('cmyk(0, 0, 100%, 0)').yellow_int8 == 255
    assert Color('cmyk(0, 0, 0, 0)').yellow_int8 == 0
    c = Color('none')
    c.yellow_int8 = 255
    assert c.yellow_int8 == 255


def test_black_int8():
    assert Color('cmyk(0, 0, 0, 100%)').black_int8 == 255
    assert Color('cmyk(0, 0, 0, 0)').black_int8 == 0
    c = Color('none')
    c.black_int8 = 255
    assert c.black_int8 == 255


def test_string():
    assert Color('black').string in ('rgb(0,0,0)', 'srgb(0,0,0)')
    assert str(Color('black')) in ('rgb(0,0,0)', 'srgb(0,0,0)')


def test_fuzz():
    c = Color('none')
    c.fuzz = 55.5
    assert c.fuzz == 55.5
    with raises(TypeError):
        c.fuzz = 'NaN'


def test_hsl():
    with Color.from_hsl(hue=0.0, saturation=1.0, lightness=0.5) as c:
        assert c == Color('#f00')
    with Color('#00f') as c:
        assert c.hsl() == (0.6666666666666666, 1.0, 0.5)


def color_memory_leak():
    for _ in range(5000):
        with Color('orange'):
            pass
    time.sleep(0.02)


@mark.skipif(memory_usage is None or MAGICK_VERSION_INFO <= (6, 6, 9, 7),
             reason='memory_usage is unavailable, or untestable')
def test_memory_leak():
    """https://github.com/emcconville/wand/pull/127"""
    minimum = 1.0
    with Color('NONE') as nil_color:
        minimum = ctypes.sizeof(nil_color.raw)
    consumes = memory_usage((color_memory_leak, (), {}))
    assert consumes[-1] - consumes[0] <= minimum


def test_color_assert_double_user_error():
    with Color('WHITE') as c:
        with raises(TypeError):
            c._assert_double('double')
        with raises(ValueError):
            c._assert_double(-1)
        with raises(ValueError):
            c._assert_double(1.1)


def test_color_assert_int8_user_error():
    with Color('WHITE') as c:
        with raises(TypeError):
            c._assert_int8(0.123)
        with raises(ValueError):
            c._assert_int8(-1)
        with raises(ValueError):
            c._assert_int8(0xFFFFF)


def test_color_assert_quantum_user_error():
    with Color('WHITE') as c:
        with raises(TypeError):
            c._assert_quantum('Q16')
        with raises(ValueError):
            c._assert_quantum(-1)


def test_color_repr_html():
    with Color('#AABBCC') as c:
        assert '#AABBCC' in c._repr_html_()



# Wand documentation build configuration file, created by
# sphinx-quickstart on Wed Sep 28 18:27:30 2011.
#
# This file is execfile()d with the current
# directory set to its containing directory.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

import sys
import os
import datetime

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.insert(0, os.path.abspath('..'))
sys.path.append(os.path.abspath('_themes'))
from wand.version import VERSION  # noqa

# Mocking C libraries to fake wand.api module which is unavailable
# on ReadTheDocs builder.
if os.environ.get('READTHEDOCS', 0):
    try:
        import wand.api  # noqa
    except ImportError:
        pass

    class Mock:
        def __init__(self, name):
            self.name = name

        def __getattr__(self, name):
            return Mock(self.name + '.' + name)

        def __repr__(self):
            return self.name

    mockapi = sys.modules['wand._api']
    mockapi.library = Mock('wand.api.library')
    mockapi.libmagick = Mock('wand.api.libmagick')
    mockapi.libc = Mock('wand.api.libc')
    sys.modules['wand'].api = sys.modules['wand.api'] = mockapi
    sys.modules['wand.version'].MAGICK_VERSION = None
    sys.modules['wand.version'].MAGICK_VERSION_NUMBER = 0x700  # Not break rtd
    sys.modules['wand.version'].MAGICK_VERSION_INFO = None
    sys.modules['wand.version'].MAGICK_RELEASE_DATE_STRING = None
    sys.modules['wand.version'].MAGICK_RELEASE_DATE = None
    sys.modules['wand.version'].QUANTUM_DEPTH = None
    sys.modules['wand.version'].QUANTUM_RANGE = None
    sys.modules['wand.version'].QUANTUM_SCALE = None
    sys.modules['wand.version'].MAGICK_HDRI = None

# -- General configuration ----------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinx.ext.autodoc', 'sphinx.ext.intersphinx',
              'sphinx.ext.extlinks', 'sphinx.ext.todo']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The encoding of source files.
# source_encoding = 'utf-8-sig'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'Wand'
copyright = str(datetime.date.today().year) + ', Hong Minhee'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = VERSION
# The full version, including alpha/beta/rc tags.
release = VERSION

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
# language = None

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
# today = ''

# Else, today_fmt is used as the format for a strftime call.
# today_fmt = '%B %d, %Y'

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
exclude_patterns = ['_build']

# The reST default role (used for this markup: `text`)
# to use for all documents.
# default_role = None

# If true, '()' will be appended to :func: etc. cross-reference text.
# add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
# add_module_names = True

# If true, sectionauthor and moduleauthor directives will be shown in the
# output. They are ignored by default.
# show_authors = False

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'trac'

# A list of ignored prefixes for module index sorting.
# modindex_common_prefix = []


# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'flask'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {'index_logo': '', 'index_logo_height': ''}

# Add any paths that contain custom themes here, relative to this directory.
html_theme_path = ['_themes']

html_style = 'wand.css'

# The name for this set of Sphinx documents.  If None, it defaults to
# "<project> v<release> documentation".
html_title = project + ' ' + release

# A shorter title for the navigation bar.  Default is the same as html_title.
html_short_title = project

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/wand.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
# html_favicon = None

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# If not '', a 'Last updated on:' timestamp is inserted at every page bottom,
# using the given strftime format.
# html_last_updated_fmt = '%b %d, %Y'

# If true, SmartyPants will be used to convert quotes and dashes to
# typographically correct entities.
html_use_smartypants = True

# Custom sidebar templates, maps document names to template names.
# html_sidebars = {}

# Additional templates that should be rendered to pages, maps page names to
# template names.
# html_additional_pages = {}

# If false, no module index is generated.
# html_domain_indices = True

# If false, no index is generated.
# html_use_index = True

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
# html_show_sourcelink = True

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
# html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
# html_show_copyright = True

# If true, an OpenSearch description file will be output, and all pages will
# contain a <link> tag referring to it.  The value of this option must be the
# base URL from which the finished HTML is served.
# html_use_opensearch = ''

# This is the file name suffix for HTML files (e.g. ".xhtml").
# html_file_suffix = None

# Output file base name for HTML help builder.
htmlhelp_basename = 'Wanddoc'


# -- Options for LaTeX output -------------------------------------------------

# The paper size ('letter' or 'a4').
# latex_paper_size = 'letter'

# The font size ('10pt', '11pt' or '12pt').
# latex_font_size = '10pt'

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author,
#  documentclass [howto/manual]).
latex_documents = [
    ('index', 'Wand.tex', 'Wand Documentation', 'Hong Minhee \\and E. McConville', 'manual'),
]

# The name of an image file (relative to this directory) to place at the top of
# the title page.
# latex_logo = None

# For "manual" documents, if this is true, then toplevel headings are parts,
# not chapters.
# latex_use_parts = False

# If true, show page references after internal links.
# latex_show_pagerefs = False

# If true, show URL addresses after external links.
# latex_show_urls = False

# Additional stuff for the LaTeX preamble.
# latex_preamble = ''

# Documents to append as an appendix to all manuals.
# latex_appendices = []

# If false, no module index is generated.
# latex_domain_indices = True


# -- Options for manual page output -------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'wand', 'Wand Documentation',
     ['Hong Minhee', 'E. McConville'], 1)
]


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'python': ('https://docs.python.org/3', None)}

extlinks = {
    'issue': ('https://github.com/emcconville/wand/issues/%s', '#%s'),
    'pull': ('https://github.com/emcconville/wand/pull/%s', '!%s'),
    'branch': ('https://github.com/emcconville/wand/compare/master...%s', '%s'),
    'commit': ('https://github.com/emcconville/wand/commit/%s', '%s'),
    'cli': ('https://imagemagick.org/script/command-line-options.php#%s', '-%s'),
}

# fall back if theme is not there
try:
    __import__('flask_theme_support')
except ImportError:
    print('-' * 74)
    print('Warning: Flask themes unavailable. Building with default theme')
    print('If you want the Flask themes, run this command and build again:')
    print('')
    print(' git submodule update --init')
    print('-' * 74)

    pygments_style = 'tango'
    html_theme = 'default'
    html_theme_options = {}
    html_logo = None
    html_style = None



import os
from pathlib import Path

from pytest import fixture, mark

from wand.version import MAGICK_VERSION, VERSION


def pytest_addoption(parser):
    parser.addoption('--skip-slow', action='store_true',
                     help='Skip slow tests')
    parser.addoption('--skip-pdf', action='store_true',
                     help='Skip any test with PDF documents.')
    parser.addoption('--skip-fft', action='store_true',
                     help='Skip any test with Forward Fourier Transform.')
    parser.addoption('--no-pdf', action='store_true',
                     help='Alias to --skip-pdf.')


def pytest_collection_modifyitems(config, items):
    skip_slow = False
    skip_pdf = False
    skip_fft = False
    if config.getoption('--skip-slow'):
        skip_slow = mark.skip('skipped; --skip-slow option was used')
    if config.getoption('--skip-pdf'):
        skip_pdf = mark.skip('skipped; --skip-pdf option was used')
    if config.getoption('--skip-fft'):
        skip_fft = mark.skip('skipped; --skip-fft option was used')
    if config.getoption('--no-pdf'):
        skip_pdf = mark.skip('skipped; --skip-pdf option was used')
    for item in items:
        if skip_slow and 'slow' in item.keywords:
            item.add_marker(skip_slow)
        if skip_pdf and 'pdf' in item.keywords:
            item.add_marker(skip_pdf)
        if skip_fft and 'fft' in item.keywords:
            item.add_marker(skip_fft)


def pytest_configure(config):
    config.addinivalue_line(
        'markers', 'slow: marks test as slow-running'
    )
    config.addinivalue_line(
        'markers', 'pdf: marks test as PDF/Ghostscript dependent'
    )
    config.addinivalue_line(
        'markers', 'fft: marks test as Forward Fourier Transform dependent'
    )


def pytest_report_header(config):
    versions = (VERSION, os.linesep, MAGICK_VERSION)
    return "Wand Version: {0}{1}ImageMagick Version: {2}".format(*versions)


@fixture
def fx_asset():
    """The fixture that provides :class:`pathlib.Path` instance that
    points the :file:`assets` directory.  You can use this in test
    functions::

        def test_something(fx_asset):
            monalisa = str(fx_asset.joinpath('mona-lisa.jpg'))
            with open(monalisa) as f:
                assert f.tell() == 0

    .. versionchanged:: 0.6.11
       Switch `py.path.local` to `pathlib.Path`.
    """
    return Path(__file__).with_name('assets')



""":mod:`wand.cdefs.magick_wand` --- Magick-Wand definitions
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. versionadded:: 0.5.0
"""
from ctypes import POINTER, c_bool, c_int, c_void_p

from wand.cdefs.wandtypes import c_ssize_t

__all__ = ('load',)


def load(lib, IM_VERSION):
    """Define Magick Wand methods. The ImageMagick version is given as
    a second argument for comparison. This will quick to determine which
    methods are available from the library, and can be implemented as::

        if IM_VERSION < 0x700:
            # ... do ImageMagick-6 methods ...
        else
            # ... do ImageMagick-7 methods ...

    .. seealso::

        #include "wand/magick-wand.h"
        // Or
        #include "MagickWand/magick-wand.h"

    :param lib: the loaded ``MagickWand`` library
    :type lib: :class:`ctypes.CDLL`
    :param IM_VERSION: the ImageMagick version number (i.e. 0x0689)
    :type IM_VERSION: :class:`numbers.Integral`

    .. versionadded:: 0.5.0

    """
    lib.ClearMagickWand.argtypes = [c_void_p]
    lib.CloneMagickWand.argtypes = [c_void_p]
    lib.CloneMagickWand.restype = c_void_p
    lib.DestroyMagickWand.argtypes = [c_void_p]
    lib.DestroyMagickWand.restype = c_void_p
    lib.IsMagickWand.argtypes = [c_void_p]
    try:
        lib.IsMagickWandInstantiated.argtypes = []
        lib.IsMagickWandInstantiated.restype = c_bool
    except AttributeError:
        lib.IsMagickWandInstantiated = None
        pass
    lib.MagickClearException.argtypes = [c_void_p]
    lib.MagickGetException.argtypes = [c_void_p, POINTER(c_int)]
    lib.MagickGetException.restype = c_void_p
    lib.MagickGetExceptionType.argtypes = [c_void_p]
    lib.MagickGetExceptionType.restype = c_int
    lib.MagickGetIteratorIndex.argtypes = [c_void_p]
    lib.MagickGetIteratorIndex.restype = c_ssize_t
    lib.MagickRelinquishMemory.argtypes = [c_void_p]
    lib.MagickRelinquishMemory.restype = c_void_p
    lib.MagickResetIterator.argtypes = [c_void_p]
    lib.MagickSetFirstIterator.argtypes = [c_void_p]
    lib.MagickSetIteratorIndex.argtypes = [c_void_p, c_ssize_t]
    lib.MagickSetIteratorIndex.restype = c_bool
    lib.MagickSetLastIterator.argtypes = [c_void_p]
    lib.MagickWandGenesis.argtypes = []
    lib.MagickWandTerminus.argtypes = []
    lib.NewMagickWandFromImage.argtypes = [c_void_p]
    lib.NewMagickWandFromImage.restype = c_void_p
    lib.NewMagickWand.argtypes = []
    lib.NewMagickWand.restype = c_void_p



from wand.exceptions import (BaseError, BaseFatalError, BaseWarning, BlobError,
                             BlobFatalError, BlobWarning, WandError,
                             WandFatalError, WandWarning)


def test_base_wand_error():
    assert issubclass(WandError, BaseError)
    assert issubclass(BlobError, BaseError)
    assert not issubclass(BlobError, WandError)


def test_base_wand_warning():
    assert issubclass(WandWarning, BaseWarning)
    assert issubclass(BlobWarning, BaseWarning)
    assert not issubclass(BlobWarning, WandWarning)


def test_base_wand_fatal_error():
    assert issubclass(WandFatalError, BaseFatalError)
    assert issubclass(BlobFatalError, BaseFatalError)
    assert not issubclass(BlobFatalError, WandFatalError)



#!/usr/bin/env python

# Python MagickWand bindings can be obtained
# from https://github.com/emcconville/wand
from wand.image import Image

import os
import re

# Android - Supporting multiple screens
# http://developer.android.com/guide/practices/screens_support.html

MANIFEST = (
    # dpi, ratio
    ('h', .75),  # high-dpi
    ('m', .5),  # medium-dpi
    ('l', .375)  # low-dpi
)

if __name__ == "__main__":
    # We start from here (2.0x size images)
    files = os.listdir("res/drawable-xhdpi/")

    # Filter out non-desired files
    files = filter(lambda x: re.match(r"^\w+(.jpg|.png)$", x) is not None,
                   files)

    for filename in files:
        with Image(filename="res/drawable-xhdpi/%s" % filename) as img:

            print("Resizing %s..." % filename)

            width, height = img.size

            for dpi, ratio in MANIFEST:
                cimg = img.clone()
                cimg.resize(int(width * ratio), int(height * ratio))
                cimg.save(filename="res/drawable-%sdpi/%s" % (dpi, filename))



#
# These test cover the Image methods that directly map to C-API function calls.
#
import io
import warnings

from pytest import mark, raises

from wand.color import Color
from wand.exceptions import MissingDelegateError, OptionError, WandRuntimeError
from wand.font import Font
from wand.image import Image
from wand.version import MAGICK_VERSION_NUMBER


def test_adaptive_blur():
    with Image(filename='rose:') as img:
        was = img.signature
        img.adaptive_blur(8, 3)
        assert was != img.signature
        was = img.signature
        img.adaptive_blur(8, 3, channel='red')
        assert was != img.signature


def test_adaptive_resize():
    with Image(filename='rose:') as img:
        img.adaptive_resize(140, 92)
        assert 140, 92 == img.size
    with Image(filename='rose:') as img:
        _, h = img.size
        img.adaptive_resize(columns=140)
        assert img.size == (140, h)
        img.adaptive_resize(rows=92)
        assert img.size == (140, 92)


@mark.skipif(MAGICK_VERSION_NUMBER < 0x690,
             reason='-adaptive-sharpen fixed in version 6.8.9-2')
def test_adaptive_sharpen():
    with Image(width=100, height=100, pseudo='plasma:') as img:
        was = img.signature
        img.adaptive_sharpen(15, 3)
        assert was != img.signature
    with Image(width=100, height=100, pseudo='plasma:') as img:
        was = img.signature
        img.adaptive_sharpen(15, 3, channel='red')
        assert was != img.signature


def test_adaptive_threshold():
    with Image(filename='rose:') as img:
        was = img.signature
        offset = 0.1 * img.quantum_range
        img.adaptive_threshold(15, 15, offset)
        assert was != img.signature


def test_affine():
    with Image(filename='rose:') as img:
        was_page = img.page
        was_sign = img.signature
        img.affine(sx=0.9, rx=1.1,
                   ry=0.1, sy=1.9,
                   tx=150, ty=-50)
        assert was_sign != img.signature
        assert was_page != img.page


def test_annotate(fx_asset):
    from wand.drawing import Drawing
    with Image(filename='rose:') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.font = str(fx_asset.joinpath('League_Gothic.otf'))
            ctx.font_size = 32
            img.annotate('Hello', ctx, left=10, baseline=img.height-10)
        assert was != img.signature
    with raises(TypeError):
        with Image(filename='rose:') as img:
            img.annotate('Hello', 0xDEADBEEF)


def test_auto_gamma():
    with Image(filename='rose:') as img:
        was = img.signature
        img.auto_gamma()
        assert was != img.signature


def test_auto_level():
    with Image(filename='rose:') as img:
        was = img.signature
        img.auto_level()
        assert was != img.signature


def test_auto_orientation(fx_asset):
    with Image(filename=str(fx_asset.joinpath('beach.jpg'))) as img:
        # if orientation is undefined nothing should be changed
        before = img[100, 100]
        img.auto_orient()
        after = img[100, 100]
        assert before == after
        assert img.orientation == 'top_left'

    fpath = str(fx_asset.joinpath('orientationtest.jpg'))
    with Image(filename=fpath) as original:
        with original.clone() as img:
            # now we should get a flipped image
            assert img.orientation == 'bottom_left'
            before = img[100, 100]
            img.auto_orient()
            after = img[100, 100]
            assert before != after
            assert img.orientation == 'top_left'

            assert img[0, 0] == original[0, -1]
            assert img[0, -1] == original[0, 0]
            assert img[-1, 0] == original[-1, -1]
            assert img[-1, -1] == original[-1, 0]


@mark.skipif(MAGICK_VERSION_NUMBER < 0x708,
             reason='Auto Threshold requires ImageMagick-7.0.8.')
def test_auto_threshold():
    with Image(filename='rose:') as img:
        was = img.signature
        img.auto_threshold()
        assert was != img.signature


@mark.skipif(MAGICK_VERSION_NUMBER < 0x711,
             reason="BilateralBlur requires ImageMagick-7.1.1")
def test_bilateral_blur():
    with Image(filename='wizard:') as img:
        was = img.signature
        img.bilateral_blur(width=8)
        assert was != img.signature


def test_black_threshold():
    with Image(filename='rose:') as img:
        was = img.signature
        img.black_threshold(Color('gray(50%)'))
        img.black_threshold('gray(50%)')
        assert was != img.signature
        with raises(TypeError):
            img.white_threshold(0xDEADBEEF)


def test_blue_shift():
    with Image(filename='rose:') as img:
        was = img.signature
        img.blue_shift(1.5)
        assert was != img.signature
        with raises(TypeError):
            img.blue_shift('NaN')


def test_blur():
    with Image(filename='rose:') as img:
        was = img.signature
        img.blur(30, 10)
        assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        img.blur(30, 10, 'red')
        assert was != img.signature


def test_border():
    with Image(filename='rose:') as img:
        was = img.signature
        with Color('red') as color:
            img.border(color, 2, 5)
            assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        img.border('pink', 2, 5)
        assert was != img.signature


def test_brightness_contrast():
    with Image(filename='rose:') as img:
        was = img.signature
        img.brightness_contrast(-10.0, 50)
        assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        img.brightness_contrast(-10.0, 50, 'red')
        assert was != img.signature


@mark.skipif(MAGICK_VERSION_NUMBER < 0x708,
             reason="Canny requires ImageMagick-7.0.8")
def test_canny():
    with Image(filename='rose:') as img:
        img.transform_colorspace('gray')
        was = img.signature
        img.canny(1, 3)
        assert was != img.signature


def test_caption(fx_asset):
    with Image(width=144, height=192, background=Color('#1e50a2')) as img:
        font = Font(
            path=str(fx_asset.joinpath('League_Gothic.otf')),
            color=Color("gold"),
            size=12,
            antialias=False
        )
        img.caption(
            'Test message',
            font=font,
            left=5, top=144,
            width=134, height=20,
            gravity='center'
        )


def test_caption_without_font():
    with Image(width=144, height=192, background=Color('#1e50a2')) as img:
        with raises(TypeError):
            img.caption(
                'Test message',
                left=5, top=144,
                width=134, height=20,
                gravity='center'
            )


def test_charcoal():
    with Image(filename='rose:') as img:
        img.transform_colorspace('gray')
        was = img.signature
        img.charcoal(1, 2)
        assert was != img.signature


def test_chop():
    with Image(filename='rose:') as img:
        w, h = img.size
        img.chop(10, 10, 10, 10)
        assert img.width == (w - 10)
        assert img.height == (h - 10)


def test_chop_gravity():
    with Image(filename='rose:') as img:
        img.chop(width=10, height=10, gravity='south_east')
        assert (60, 36, 0, 0) == img.page
    with Image(filename='rose:') as img:
        img.chop(width=10, height=10, x=10, y=10, gravity='north')
        assert (60, 36, 0, 0) == img.page


@mark.skipif(MAGICK_VERSION_NUMBER < 0x709,
             reason="Clahe requires Imagemagick-7.0.9.")
def test_clahe():
    with Image(filename='rose:') as img:
        was = img.signature
        img.clahe(10, 10, 128, 3)
        assert was != img.signature


def test_clamp():
    with Image(filename='rose:') as img:
        assert img.clamp(channel='red')


def test_clut():
    with Image(filename='rose:') as img:
        was = img.signature
        with Image(img) as clut:
            clut.unique_colors()
            clut.flop()
            img.clut(clut)
            assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        with Image(img) as clut:
            clut.unique_colors()
            clut.flop()
            img.clut(clut, channel='red')
            assert was != img.signature
        with raises(TypeError):
            img.clut(0xDEADBEEF)


def test_coalesce(fx_asset):
    with Image(filename=str(fx_asset.joinpath('nocomments.gif'))) as img1:
        with Image(img1) as img2:
            img2.coalesce()
            assert img1.signature != img2.signature
            assert img1.size == img2.size


def test_color_decision_list():
    ccc = """
    <ColorCorrectionCollection xmlns="urn:ASC:CDL:v1.2">
        <ColorCorrection id="cc03345">
            <SOPNode>
                <Slope> 0.9 1.2 0.5 </Slope>
                <Offset> 0.4 -0.5 0.6 </Offset>
                <Power> 1.0 0.8 1.5 </Power>
            </SOPNode>
            <SATNode>
                <Saturation> 0.85 </Saturation>
            </SATNode>
        </ColorCorrection>
    </ColorCorrectionCollection>
    """
    with Image(filename='rose:') as img:
        was = img.signature
        assert img.color_decision_list(ccc)
        assert was != img.signature
    with Image(filename='rose:') as img:
        assert img.cdl(ccc)


def test_color_map(fx_asset):
    with Image(filename=str(fx_asset.joinpath('trim-color-test.png'))) as img:
        img.type = 'palette'
        assert img[0, 0] == img.color_map(0)
        orange = Color('ORANGE')
        assert orange == img.color_map(0, orange)
        assert orange == img.color_map(0, 'ORANGE')
        assert orange == img.color_map(0)
        with raises(TypeError):
            img.color_map('NaN')
        with raises(TypeError):
            img.color_map(0, 0xDEADBEEF)
        with raises(ValueError):
            img.color_map(-1, orange)


def test_color_matrix():
    with Image(filename='rose:') as img:
        was = img.signature
        matrix = [
            [0.9, 0.0, 0.0],
            [0.0, 0.9, 0.0],
            [0.0, 0.0, 1.3]
        ]
        img.color_matrix(matrix)
        assert was != img.signature
        with raises(TypeError):
            img.color_matrix(0xDEADBEEF)
        with raises(TypeError):
            img.color_matrix((0, 0, 0, 0, 0))
        with raises(ValueError):
            img.color_matrix([[1, 0], [0, 1, 0]])


@mark.skipif(MAGICK_VERSION_NUMBER < 0x70A,
             reason='Color Threshold requires ImageMagick-7.0.10')
def test_color_threshold():
    with Image(width=100, height=100, pseudo='plasma:') as img:
        was = img.signature
        img.color_threshold(start='#000', stop='#ccc')
        assert was != img.signature


def test_colorize():
    with Image(filename='rose:') as img:
        was = img.signature
        img.colorize('blue', 'blue')
        assert was != img.signature
        with raises(TypeError):
            img.colorize(0xDEADBEEF, Color('blue'))
        with raises(TypeError):
            img.colorize(Color('blue'), 0xDEADBEEF)


def test_combine():
    with Image() as img:
        img.pseudo(width=1, height=1, pseudo='xc:grey100')
        img.pseudo(width=1, height=1, pseudo='xc:grey50')
        img.pseudo(width=1, height=1, pseudo='xc:grey0')
        assert len(img.sequence) == 3
        img.colorspace = 'rgb'
        img.combine()
        assert len(img.sequence) == 1
        pixel = img[0, 0]
        assert pixel.red > pixel.green > pixel.blue


def test_compare():
    with Image(filename="rose:") as img1:
        with img1.clone() as img2:
            img2.import_pixels(10, 10, 1, 1, 'R', 'char', b'\00')
            cmp_img, err = img1.compare(img2, 'absolute',
                                        highlight=Color('orange'),
                                        lowlight=Color('gray90'))
            assert cmp_img
            assert err >= 0.0


@mark.skipif(MAGICK_VERSION_NUMBER < 0x708,
             reason="Complex requires ImageMagick-7.0.8.")
def test_complex():
    with Image(width=1, height=1, pseudo='xc:gray25') as a:
        with Image(width=1, height=1, pseudo='xc:gray50') as b:
            a.image_add(b)
            a.iterator_reset()
            with a.complex('add') as img:
                assert a.signature != img.signature


def test_composite():
    with Image(filename='rose:') as img:
        was = img.signature
        with Image(filename='rose:') as fg:
            img.composite(fg, 5, 10)
        assert img.signature != was


def test_composite_arguments():
    with Image(filename='rose:') as img:
        base = img.signature
        left = base
        right = base
        with img.clone() as img1:
            with Image(width=img.width,
                       height=img.height,
                       pseudo='gradient:') as mask:
                img1.composite(mask, operator='blend')
            left = img1.signature
            assert base != left
        with img.clone() as img2:
            with Image(width=img.width,
                       height=img.height,
                       pseudo='gradient:') as mask:
                img2.composite(mask, operator='blend', arguments='7,7')
            right = img2.signature
            assert base != right
        assert left != right


def test_composite_gravity():
    green = Color('GREEN')
    red = Color('RED')
    with Image(width=100, height=100, background=green) as src:
        with Image(width=10, height=10, background=red) as dst:
            src.composite(dst, gravity='east')
        assert src[0, 50] == green
        assert src[99, 50] == red


def test_composite_channel(fx_asset):
    with Image(filename=str(fx_asset.joinpath('beach.jpg'))) as orig:
        w, h = orig.size
        left = w // 4
        top = h // 4
        right = left * 3 - 1
        bottom = h // 4 * 3 - 1
        # List of (x, y) points that shouldn't be changed:
        outer_points = [
            (0, 0), (0, h - 1), (w - 1, 0), (w - 1, h - 1),
            (left, top - 1), (left - 1, top), (left - 1, top - 1),
            (right, top - 1), (right + 1, top), (right + 1, top - 1),
            (left, bottom + 1), (left - 1, bottom), (left - 1, bottom + 1),
            (right, bottom + 1), (right + 1, bottom), (right + 1, bottom + 1)
        ]
        if MAGICK_VERSION_NUMBER < 0x700:
            channel_name = 'red'
        else:
            channel_name = 'default_channels'
        with orig.clone() as img:
            with Color('black') as color:
                with Image(width=w // 2, height=h // 2,
                           background=color) as cimg:
                    img.composite_channel(channel_name, cimg, 'copy_red',
                                          w // 4, h // 4)
            # These points should be not changed:
            for point in outer_points:
                assert orig[point] == img[point]
            # Inner pixels should lost its red color (red becomes 0)
            for point in zip([left, right], [top, bottom]):
                with orig[point] as oc:
                    with img[point] as ic:
                        assert not ic.red
                        assert ic.green == oc.green
                        assert ic.blue == oc.blue


def test_composite_channel_arguments():
    channel_name = 'default_channels'
    with Image(filename='rose:') as img:
        base = img.signature
        left = base
        right = base
        with img.clone() as img1:
            with Image(width=img.width,
                       height=img.height,
                       pseudo='gradient:') as mask:
                img1.composite_channel(channel_name, mask, 'blend')
            left = img1.signature
            assert base != left
        with img.clone() as img2:
            with Image(width=img.width,
                       height=img.height,
                       pseudo='gradient:') as mask:
                img2.composite_channel(channel_name, mask, 'blend',
                                       arguments='7,7')
            right = img2.signature
            assert base != right
        assert left != right


def test_composite_channel_gravity():
    green = Color('GREEN')
    red = Color('RED')
    channel_name = 'default_channels'
    with Image(width=100, height=100, background=green) as src:
        with Image(width=10, height=10, background=red) as dst:
            src.composite_channel(channel_name, dst, 'over', gravity='east')
        assert src[0, 50] == green
        assert src[99, 50] == red


def test_concat():
    with Image(filename='rose:') as img:
        img.read(filename='rose:')
        with Image(img) as row:
            row.concat()
            assert row.size == (140, 46)
        with Image(img) as row:
            row.concat(True)
            assert row.size == (70, 92)


@mark.skipif(MAGICK_VERSION_NUMBER < 0x708,
             reason='Connected Components requires ImageMagick-7.0.8.')
def test_connected_components(fx_asset):
    with Image(filename=str(fx_asset.joinpath('ccobject.png'))) as img:
        objects = img.connected_components()
        assert 2 == len(objects)


def test_contrast():
    with Image(filename='rose:') as img:
        was = img.signature
        img.contrast()
        assert was != img.signature


def test_contrast_stretch():
    with Image(filename='rose:') as img:
        was = img.signature
        img.contrast_stretch(0.15)
        assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        img.contrast_stretch(0.15, channel='red')
        assert was != img.signature


def test_contrast_stretch_user_error():
    with Image(filename='rose:') as img:
        with raises(TypeError):
            img.contrast_stretch('NaN')
        with raises(TypeError):
            img.contrast_stretch(0.1, 'NaN')
        with raises(ValueError):
            img.contrast_stretch(0.1, channel='Not a channel')


@mark.skipif(MAGICK_VERSION_NUMBER < 0x70A,
             reason='Convex Hull requires ImageMagick-7.0.10.')
def test_convex_hull(fx_asset):
    fpath = str(fx_asset.joinpath('horizon_sunset_border2.jpg'))
    with Image(filename=fpath) as img:
        points = img.convex_hull(background='black')
        assert len(points) > 0


def test_crop(fx_asset):
    """Crops in-place."""
    with Image(filename=str(fx_asset.joinpath('croptest.png'))) as img:
        with img.clone() as cropped:
            assert cropped.size == img.size
            cropped.crop(100, 100, 200, 200)
            assert cropped.size == (100, 100)
            with Color('#000') as black:
                for row in cropped:
                    for col in row:
                        assert col == black
        with img.clone() as cropped:
            assert cropped.size == img.size
            cropped.crop(100, 100, width=100, height=100)
            assert cropped.size == (100, 100)
        with img.clone() as cropped:
            assert cropped.size == img.size
            cropped.crop(left=150, bottom=150)
            assert cropped.size == (150, 150)
        with img.clone() as cropped:
            assert cropped.size == img.size
            cropped.crop(left=150, height=150)
            assert cropped.size == (150, 150)
        with img.clone() as cropped:
            assert cropped.size == img.size
            cropped.crop(-200, -200, -100, -100)
            assert cropped.size == (100, 100)
        with img.clone() as cropped:
            assert cropped.size == img.size
            cropped.crop(top=100, bottom=200)
            assert cropped.size == (300, 100)
        with img.clone() as cropped:
            width, height = img.size
            assert cropped.crop(left=0, top=0, width=width, height=height)
        with raises(ValueError):
            img.crop(0, 0, 500, 500)
        with raises(ValueError):
            img.crop(290, 290, 50, 50)
        with raises(ValueError):
            img.crop(290, 290, width=0, height=0)


def test_crop_error(fx_asset):
    """Crop errors."""
    with Image(filename=str(fx_asset.joinpath('croptest.png'))) as img:
        with raises(TypeError):
            img.crop(right=1, width=2)
        with raises(TypeError):
            img.crop(bottom=1, height=2)


def test_crop_gif(tmp_path, fx_asset):
    fpath = str(fx_asset.joinpath('nocomments-delay-100.gif'))
    with Image(filename=fpath) as img:
        with img.clone() as d:
            assert d.size == (350, 197)
            for s in d.sequence:
                assert s.delay == 100
            d.crop(50, 50, 200, 150)
            d.save(filename=str(tmp_path / '50_50_200_150.gif'))
        with Image(filename=str(tmp_path / '50_50_200_150.gif')) as d:
            assert len(d.sequence) == 46
            assert d.size == (150, 100)
            for s in d.sequence:
                assert s.delay == 100


def test_crop_gravity(fx_asset):
    with Image(filename=str(fx_asset.joinpath('croptest.png'))) as img:
        width = int(img.width / 3)
        height = int(img.height / 3)
        mid_width = int(width / 2)
        mid_height = int(height / 2)
        with img.clone() as center:
            center.crop(width=width, height=height, gravity='center')
            assert center[mid_width, mid_height] == Color('black')
        with img.clone() as northwest:
            northwest.crop(width=width, height=height, gravity='north_west')
            assert northwest[mid_width, mid_height] == Color('transparent')
        with img.clone() as southeast:
            southeast.crop(width=width, height=height, gravity='south_east')
            assert southeast[mid_width, mid_height] == Color('transparent')


def test_crop_issue367():
    with Image(filename='rose:') as img:
        expected = img.size
        for gravity in ('north_west', 'north', 'north_east',
                        'west', 'center', 'east',
                        'south_west', 'south', 'south_east',):
            with Image(img) as actual:
                actual.crop(width=200, height=200, gravity=gravity)
                assert actual.size == expected


def test_crop_issue669():
    with Image(filename='rose:') as img:
        img.crop(width=50, height=25, left=10, gravity='south')
        assert 50, 25 == img.size


def test_cycle_color_map(fx_asset):
    with Image(filename=str(fx_asset.joinpath('trim-color-test.png'))) as img:
        img.type = 'palette'
        lime = img[0, 0]
        img.cycle_color_map(1)
        assert img[-1, 0] == lime
        img.cycle_color_map(-1)
        assert img[0, 0] == lime
        with raises(TypeError):
            img.cycle_color_map('NaN')


def test_deskew():
    with Image(filename='rose:') as img:
        was = img.signature
        img.deskew(0.4 * img.quantum_range)
        assert was != img.signature


def test_despeckle():
    with Image(filename='rose:') as img:
        was = img.signature
        img.despeckle()
        assert was != img.signature


def test_distort():
    """Distort image."""
    with Image(filename='rose:') as img:
        was = img.signature
        w, h = img.size
        img.distort('resize', (w*2, h*2))
        neu = img.signature
        assert was != neu
    # Let's ensure a distorted image with a defined filter doesn't
    # match default.
    with Image(filename='rose:') as img:
        was = img.signature
        w, h = img.size
        img.distort('resize', (w*2, h*2), filter='lanczos')
        assert neu != img.signature


def test_distort_error():
    """Distort image with user error"""
    with Image(filename='rose:') as img:
        with raises(ValueError):
            img.distort('mirror', (1,))
        with raises(TypeError):
            img.distort('perspective', 1)


def test_edge():
    with Image(filename='rose:') as img:
        was = img.signature
        img.edge(1.5)
        assert was != img.signature


def test_emboss():
    with Image(filename='rose:') as img:
        was = img.signature
        img.emboss(1.5, 0.25)
        assert was != img.signature


def test_encipher_decipher():
    with Image(filename='rose:') as img:
        img.depth = 8  # Safety
        was = img.signature
        img.encipher(passphrase='secret')
        assert was != img.signature
        img.decipher(passphrase='secret')
        assert was == img.signature


def test_enhance():
    with Image(filename='rose:') as img:
        was = img.signature
        img.enhance()
        assert was != img.signature


def test_equalize():
    with Image(filename='rose:') as img:
        was = img.signature
        img.equalize()
        assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        img.equalize(channel='red')
        assert was != img.signature


def test_evaluate():
    with Image(filename='hald:3') as img:
        with img.clone() as percent_img:
            fifty_percent = percent_img.quantum_range * 0.5
            percent_img.evaluate('set', fifty_percent)
            with percent_img[10, 10] as gray:
                assert abs(gray.red - Color('gray50').red) < 0.01
        with img.clone() as literal_img:
            literal_img.evaluate('divide', 2, channel='red')
            with img[0, 0] as org_color:
                expected_color = (org_color.red_int8 * 0.5)
                with literal_img[0, 0] as actual_color:
                    assert abs(expected_color - actual_color.red_int8) < 1


def test_evaluate_user_error():
    with Image(filename='hald:3') as img:
        with raises(ValueError):
            img.evaluate(operator='Nothing')
        with raises(TypeError):
            img.evaluate(operator='set', value='NaN')
        with raises(ValueError):
            img.evaluate(operator='set', value=1.0, channel='Not a channel')


def test_evaluate_images():
    with Image(filename='hald:3') as img:
        with img.clone() as i:
            img.image_add(i)
        with img.evaluate_images(operator='add') as nue:
            assert nue != img


def test_export_pixels():
    with Image(filename='hald:2') as img:
        img.depth = 8  # Not need, but want to match import.
        data = img.export_pixels(x=0, y=0, width=4, height=1,
                                 channel_map='RGBA', storage='char')
        expected = [0x00, 0x00, 0x00, 0xFF,
                    0x55, 0x00, 0x00, 0xFF,
                    0xAA, 0x00, 0x00, 0xFF,
                    0xFF, 0x00, 0x00, 0xFF]
        assert data == expected
        # Test Bad value
        with raises(TypeError):
            img.export_pixels(x='NaN')
        with raises(TypeError):
            img.export_pixels(y='NaN')
        with raises(TypeError):
            img.export_pixels(width='NaN')
        with raises(TypeError):
            img.export_pixels(height='NaN')
        with raises(TypeError):
            img.export_pixels(channel_map=0xDEADBEEF)
        with raises(ValueError):
            img.export_pixels(channel_map='NaN')
        with raises(ValueError):
            img.export_pixels(storage='NaN')


def test_export_pixels_issue_413():
    x = 10
    y = 10
    width = 20
    height = 10
    with Image(width=50, height=50, background=Color('GREEN')) as img:
        export = img.export_pixels(x=x, y=y,
                                   width=width, height=height)
        assert export
        export = img.export_pixels(x=x, y=y,
                                   width=width + x, height=height + y)
        assert export


def test_extent():
    with Image(filename='rose:') as img:
        with img.clone() as extended:
            assert extended.size == img.size
            extended.extent(width=500)
            assert extended.width == 500
            assert extended.height == img.height

        with img.clone() as extended:
            assert extended.size == img.size
            extended.extent(height=500)
            assert extended.width == img.width
            assert extended.height == 500

        with raises(ValueError):
            img.extent(width=-10)

        with raises(ValueError):
            img.extent(height=-10)


def test_extent_gravity():
    with Image(filename='rose:') as img:
        img.extent(width=10, height=10, gravity='south_east')
        assert (10, 10, 0, 0) == img.page
        img.extent(width=100, height=100, gravity='center')
        assert (100, 100, 0, 0) == img.page
    with Image(filename='rose:') as img:
        img.extent(x=10, gravity='north')
        assert 70, 46 == img.size


def test_features():
    with Image(filename='rose:') as img:
        cf = img.features(10)
        assert 'red' in cf
        assert 'contrast' in cf['red']
        assert 'horizontal' in cf['red']['contrast']


def test_flip():
    with Image(filename='rose:') as img:
        with img.clone() as flipped:
            flipped.flip()
            assert flipped[0, 0] == img[0, -1]
            assert flipped[0, -1] == img[0, 0]
            assert flipped[-1, 0] == img[-1, -1]
            assert flipped[-1, -1] == img[-1, 0]


def test_floodfill():
    with Image(filename='wizard:') as img:
        was = img.signature
        img.floodfill(fill='green', fuzz=0.2)
        assert was != img.signature


def test_flop():
    with Image(filename='rose:') as img:
        with img.clone() as flopped:
            flopped.flop()
            assert flopped[0, 0] == img[-1, 0]
            assert flopped[-1, 0] == img[0, 0]
            assert flopped[0, -1] == img[-1, -1]
            assert flopped[-1, -1] == img[0, -1]


@mark.fft
def test_forward_fourier_transform():
    with Image(filename='rose:') as img:
        was = img.signature
        img.forward_fourier_transform()
        assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        img.fft()
        assert was != img.signature


def test_frame():
    with Image(filename='rose:') as img:
        img.frame(width=4, height=4)
        assert img[0, 0] == img[-1, -1]
        assert img[-1, 0] == img[0, -1]
    with Color('green') as green:
        with Image(filename='rose:') as img:
            img.frame(matte=green, width=2, height=2)
            assert img[0, 0] == green
            assert img[-1, -1] == green
        with Image(filename='rose:') as img:
            img.frame(matte='green', width=2, height=2)
            assert img[0, 0] == green
            assert img[-1, -1] == green


def test_frame_error():
    with Image(filename='rose:') as img:
        with raises(TypeError):
            img.frame(width='one')
        with raises(TypeError):
            img.frame(height=3.5)
        with raises(TypeError):
            img.frame(inner_bevel=None)
        with raises(TypeError):
            img.frame(outer_bevel='large')


def test_function():
    with Image(filename='wizard:') as img:
        was = img.signature
        img.function(function='polynomial',
                     arguments=(4, -4, 1))
        assert was != img.signature
        was = img.signature
        img.function(function='sinusoid',
                     arguments=(1,),
                     channel='red')
        assert was != img.signature


def test_function_error():
    with Image(filename='rose:') as img:
        with raises(ValueError):
            img.function('bad function', 1)
        with raises(TypeError):
            img.function('sinusoid', 1)
        with raises(ValueError):
            img.function('sinusoid', (1,), channel='bad channel')


def test_fx():
    with Image(width=2, height=2, background=Color('black')) as xc1:
        # NavyBlue == #000080
        with xc1.fx('0.5019', channel='blue') as xc2:
            assert abs(xc2[0, 0].blue - Color('navy').blue) < 0.0001

    with Image(width=2, height=1, background=Color('white')) as xc1:
        with xc1.fx('0') as xc2:
            assert xc2[0, 0].red == 0


def test_fx_error():
    with Image() as empty_wand:
        with raises(WandRuntimeError):
            with empty_wand.fx('8') as _:
                pass
    with Image(filename='rose:') as xc:
        with raises(OptionError):
            with xc.fx('NULL'):
                pass
        with raises(TypeError):
            with xc.fx(('p[0,0]',)):
                pass
        with raises(TypeError):
            with xc.fx('p[0,0]', True):
                pass


def test_gamma():
    # Value under 1.0 is darker, and above 1.0 is lighter
    middle_point = 35, 23
    with Image(filename='rose:') as img:
        with img.clone() as lighter:
            lighter.gamma(1.5)
            assert img[middle_point].red < lighter[middle_point].red
        with img.clone() as darker:
            darker.gamma(0.5)
            assert img[middle_point].red > darker[middle_point].red


def test_gamma_channel():
    # Value under 1.0 is darker, and above 1.0 is lighter
    middle_point = 35, 23
    with Image(filename='rose:') as img:
        with img.clone() as lighter:
            lighter.gamma(1.5, channel='red')
            assert img[middle_point].red < lighter[middle_point].red
        with img.clone() as darker:
            darker.gamma(0.5, channel='red')
            assert img[middle_point].red > darker[middle_point].red


def test_gamma_user_error():
    with Image(filename='rose:') as img:
        with raises(TypeError):
            img.gamma('NaN;')
        with raises(ValueError):
            img.gamma(0.0, 'no channel')


def test_gaussian_blur():
    with Image(filename='rose:') as img:
        was = img.signature
        img.gaussian_blur(30, 10)
        assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        img.gaussian_blur(30, 10, channel='red')
        assert was != img.signature


def test_get_image_distortion():
    with Image(filename='wizard:') as orig:
        with orig.clone() as img:
            img.implode(0.5)
            err = orig.get_image_distortion(img, 'absolute')
            assert err >= 0.0


def test_hald_clut():
    with Image(filename='rose:') as img:
        was = img.signature
        with Image(filename='hald:3') as hald:
            hald.gamma(0.367)
            img.hald_clut(hald)
        assert was != img.signature
        was = img.signature
        with Image(filename='hald:3') as hald:
            hald.gamma(0.367)
            img.hald_clut(hald, channel='red')
        assert was != img.signature
        with raises(TypeError):
            img.hald_clut(0xDEADBEEF)


@mark.skipif(MAGICK_VERSION_NUMBER < 0x708,
             reason='Hough Lines requires ImageMagick-7.0.8.')
def test_hough_lines(fx_asset):
    with Image(filename=str(fx_asset.joinpath('ccobject.png'))) as img:
        was = img.signature
        img.hough_lines(width=3, height=3)
        assert was != img.signature


def test_implode():
    with Image(filename='rose:') as img:
        was = img.signature
        img.implode(amount=1.0)
        assert was != img.signature


def test_import_pixels():
    data = [0xFF, 0x00, 0x00, 0xFF,
            0x00, 0xFF, 0x00, 0xFF,
            0x00, 0x00, 0xFF, 0xFF,
            0x00, 0x00, 0x00, 0x00]
    with Image(width=4, height=1, background=Color('BLACK')) as dst:
        dst.depth = 8  # For safety
        was = dst.signature
        dst.import_pixels(x=0, y=0, width=4, height=1,
                          channel_map='RGBA', storage='char',
                          data=data)
        assert was != dst.signature
        with raises(TypeError):
            dst.import_pixels(x='NaN')
        with raises(TypeError):
            dst.import_pixels(y='NaN')
        with raises(TypeError):
            dst.import_pixels(width='NaN')
        with raises(TypeError):
            dst.import_pixels(height='NaN')
        with raises(TypeError):
            dst.import_pixels(channel_map=0xDEADBEEF)
        with raises(ValueError):
            dst.import_pixels(channel_map='NaN')
        with raises(ValueError):
            dst.import_pixels(storage='NaN')
        with raises(TypeError):
            dst.import_pixels(data=0xDEADBEEF)
        with raises(ValueError):
            dst.import_pixels(data=[0x00, 0xFF])


def test_import_pixels_issue_413():
    x = 10
    y = 10
    width = 20
    height = 10
    with Image(width=50, height=50, background=Color('GREEN')) as img:
        blank = [0xFF] * 600
        img.import_pixels(x=x, y=y,
                          width=width, height=height,
                          channel_map='RGB',
                          data=blank)
        assert img
        blank = [0xFF] * 1800
        img.import_pixels(x=x, y=y,
                          width=width + x, height=height + y,
                          channel_map='RGB',
                          data=blank)
        assert img


@mark.fft
def test_inverse_fourier_transform(fx_asset):
    magnitude = str(fx_asset.joinpath('ccobject_magnitude.png'))
    phase = str(fx_asset.joinpath('ccobject_phase.png'))
    with Image(filename=magnitude) as a:
        was = a.signature
        with Image(filename=phase) as b:
            a.inverse_fourier_transform(b)
        assert was != a.signature
    with Image(filename=magnitude) as a:
        with Image(filename=phase) as b:
            assert a.ift(b)
    with raises(TypeError):
        with Image(filename=magnitude) as a:
            a.inverse_fourier_transform(0xDEADBEEF)


def test_iterator(fx_asset):
    with Image(filename=str(fx_asset.joinpath('animation.gif'))) as img:
        length = img.iterator_length()
        assert length == 4
        img.iterator_reset()
        idx = img.iterator_get()
        assert idx == 0
        while img.iterator_next():
            pass
        idx = img.iterator_get()
        assert idx == 3
        img.iterator_first()
        idx = img.iterator_get()
        assert idx == 0
        img.iterator_last()
        while img.iterator_previous():
            pass
        idx = img.iterator_get()
        assert idx == 0
        img.iterator_set(2)
        idx = img.iterator_get()
        assert idx == 2


@mark.skipif(MAGICK_VERSION_NUMBER < 0x70B,
             reason='Kmeans requires ImageMagick-7.0.11')
def test_kmeans():
    with Image(filename='rose:') as img:
        was = img.signature
        img.kmeans(64)
        assert was != img.signature


def test_kurtosis_channel():
    with Image(filename='rose:') as img:
        r = img.kurtosis_channel('red')
        assert len(r) == 2
        with raises(ValueError):
            img.kurtosis_channel('unknown')


@mark.skipif(MAGICK_VERSION_NUMBER < 0x708,
             reason='Kuwahara requires ImageMagick-7.0.8.')
def test_kuwahara():
    with Image(filename='rose:') as img:
        was = img.signature
        img.kuwahara(3.0)
        assert was != img.signature


def test_label(fx_asset):
    font_path = str(fx_asset.joinpath('League_Gothic.otf'))
    with Image(filename='rose:') as img:
        was = img.signature
        img.label('a', left=0, top=0, font=Font(font_path, 12))
        now = img.signature
        assert now != was
        img.label('b', font=Font(font_path, 12), gravity='south')
        assert img.signature != now
    with raises(TypeError):
        with Image(filename='rose:') as img:
            img.label('x')


def test_level():
    with Image(width=100, height=100, pseudo='gradient:') as img:
        # Adjust the levels to make this image entirely black
        was = img.signature
        img.level(black=0.99, white=1.0)
        assert was != img.signature
    with Image(width=100, height=100, pseudo='gradient:') as img:
        # Adjust the levels to make this image entirely white
        was = img.signature
        img.level(0, 0.01)
        assert was != img.signature
    with Image(width=100, height=100, pseudo='gradient:') as img:
        # Adjust the image's gamma to darken its midtones
        was = img.signature
        img.level(gamma=0.5)
        assert was != img.signature
    with Image(width=100, height=100, pseudo='gradient:') as img:
        # Adjust the image's gamma to lighten its midtones
        was = img.signature
        img.level(0, 1, 2.5)
        assert was != img.signature


def test_level_channel():
    with Image(width=100, height=100, pseudo='gradient:') as img:
        # Adjust each channel level to make it entirely black
        was = img.signature
        img.level(0.99, 1.0, channel='red')
        assert was != img.signature
    with Image(width=100, height=100, pseudo='gradient:') as img:
        # Adjust each channel level to make it entirely white
        was = img.signature
        img.level(0.0, 0.01, channel='green')
        assert was != img.signature
    with Image(width=100, height=100, pseudo='gradient:') as img:
        # Adjust each channel's gamma to darken its midtones
        was = img.signature
        img.level(gamma=0.5, channel='blue')
        assert was != img.signature
    with Image(width=100, height=100, pseudo='gradient:') as img:
        # Adjust each channel's gamma to lighten its midtones
        was = img.signature
        img.level(0, 1, 2.5, 'red')
        assert was != img.signature


def test_level_user_error():
    with Image(width=100, height=100, pseudo='gradient:') as img:
        with raises(TypeError):
            img.level(black='NaN')
        with raises(TypeError):
            img.level(white='NaN')
        with raises(TypeError):
            img.level(gamma='NaN')
        with raises(ValueError):
            img.level(channel='404')


@mark.skipif(MAGICK_VERSION_NUMBER < 0x708,
             reason='Levelize requires ImageMagick-7.0.8.')
def test_levelize():
    with Image(filename='rose:') as img:
        was = img.signature
        img.levelize(3.0)
        assert was != img.signature


def test_linear_stretch():
    with Image(width=100, height=100, pseudo='gradient:') as img:
        was = img.signature
        img.linear_stretch(black_point=0.15,
                           white_point=0.15)
        assert was != img.signature


def test_linear_stretch_user_error():
    with Image(width=100, height=100, pseudo='gradient:') as img:
        with raises(TypeError):
            img.linear_stretch(white_point='NaN',
                               black_point=0.5)
        with raises(TypeError):
            img.linear_stretch(white_point=0.5,
                               black_point='NaN')


def test_liquid_rescale():
    with Image(filename='wizard:') as orig:
        with orig.clone() as img:
            try:
                img.liquid_rescale(600, 600)
            except MissingDelegateError:
                warnings.warn('skip liquid_rescale test; has no LQR delegate')
            else:
                assert img.size == (600, 600)


@mark.skipif(MAGICK_VERSION_NUMBER < 0x693,
             reason='Local Contrast not supported.')
def test_local_contrast():
    with Image(filename='rose:') as img:
        was = img.signature
        img.local_contrast()
        assert was != img.signature


def test_mean_channel():
    with Image(filename='rose:') as img:
        r = img.mean_channel('red')
        assert len(r) == 2
        with raises(ValueError):
            img.mean_channel('unknown')


def test_magnify():
    with Image(filename='rose:') as img:
        expected = img.width * 2
        img.magnify()
        assert expected == img.width


@mark.skipif(MAGICK_VERSION_NUMBER < 0x708,
             reason='Mean Shift requires ImageMagick-7.0.8.')
def test_mean_shift():
    with Image(filename='rose:') as img:
        was = img.signature
        img.mean_shift(width=5, height=5)
        assert was != img.signature


def test_merge_layers():
    for method in ['merge', 'flatten', 'mosaic']:
        with Image(filename='wizard:') as img1:
            with Image(filename='logo:') as img2:
                img1.sequence.append(img2)
                assert len(img1.sequence) == 2
                img1.merge_layers(method)
                assert len(img1.sequence) == 1


def test_merge_layers_bad_method():
    with Image(filename='rose:') as img:
        for method in ('', 'mosaic' 'junk'):
            with raises(ValueError):
                img.merge_layers(method)
        with raises(TypeError):
            img.merge_layers(None)


def test_merge_layers_method_flatten():
    with Image(width=16, height=16) as img1:
        img1.background_color = Color('black')
        img1.alpha_channel = False
        with Image(width=32, height=32) as img2:
            img2.background_color = Color('white')
            img2.alpha_channel = False
            img2.transform(crop='16x16+8+8')
            img1.sequence.append(img2)
            img1.merge_layers('flatten')
            assert img1.size == (16, 16)


def test_merge_layers_method_merge():
    with Image(width=16, height=16) as img1:
        img1.background_color = Color('black')
        img1.alpha_channel = False
        with Image(width=32, height=32) as img2:
            img2.background_color = Color('white')
            img2.alpha_channel = False
            img2.transform(crop='16x16+8+8')
            img1.sequence.append(img2)
            img1.merge_layers('merge')
            assert img1.size == (24, 24)


def test_merge_layers_method_merge_neg_offset():
    with Image(width=16, height=16) as img1:
        img1.background_color = Color('black')
        img1.alpha_channel = False
        with Image(width=16, height=16) as img2:
            img2.background_color = Color('white')
            img2.alpha_channel = False
            img2.page = (16, 16, -8, -8)
            img1.sequence.append(img2)
            img1.merge_layers('merge')
            assert img1.size == (24, 24)


def test_merge_layers_method_mosaic():
    with Image(width=16, height=16) as img1:
        img1.background_color = Color('black')
        img1.alpha_channel = False
        with Image(width=32, height=32) as img2:
            img2.background_color = Color('white')
            img2.alpha_channel = False
            img2.transform(crop='16x16+8+8')
            img1.sequence.append(img2)
            img1.merge_layers('mosaic')
            assert img1.size == (24, 24)


def test_merge_layers_method_mosaic_neg_offset():
    with Image(width=16, height=16) as img1:
        img1.background_color = Color('black')
        img1.alpha_channel = False
        with Image(width=16, height=16) as img2:
            img2.background_color = Color('white')
            img2.alpha_channel = False
            img2.page = (16, 16, -8, -8)
            img1.sequence.append(img2)
            img1.merge_layers('mosaic')
            assert img1.size == (16, 16)


@mark.skipif(MAGICK_VERSION_NUMBER < 0x70A,
             reason='Minimum Bounding Box requires ImageMagick-7.0.10.')
def test_minimum_bounding_box():
    with Image(filename='wizard:') as img:
        img.fuzz = 0.1 * img.quantum_range
        img.background_color = 'white'
        mbr = img.minimum_bounding_box()
        assert img.width > mbr.get('width', img.width)
        assert img.height > mbr.get('height', img.height)


def test_mode():
    with Image(filename='rose:') as img:
        was = img.signature
        img.mode(5)
        assert was != img.signature


def test_modulate():
    with Image(filename='rose:') as img:
        was = img.signature
        img.modulate(120, 120, 120)
        assert was != img.signature


def test_morph():
    with Image(filename='rose:') as img:
        img.read(filename='rose:')
        with img.morph(4) as morph:
            assert morph.iterator_length() != img.iterator_length()


def test_morphology_builtin():
    known = []
    args = (('erode', 'ring'),
            ('dilate', 'disk:5'),
            ('open', 'octagon'),
            ('smooth', 'rectangle:0x-1'),
            ('thinning', 'edges'),
            ('distance', 'euclidean:4,10!'),
            ('thicken', 'unity:x5'),
            ('close', 'manhattan:20x25%'),
            ('hit_and_miss', 'chebyshev:5.0'))
    for arg in args:
        with Image(filename='rose:') as img:
            img.morphology(*arg)
            assert img.signature not in known
            known.append(img.signature)
    with Image(filename='rose:') as img:
        with raises(TypeError):
            img.morphology(method=0xDEADBEEF)
        with raises(TypeError):
            img.morphology(method='close',
                           kernel=0xDEADBEEF)
        with raises(TypeError):
            img.morphology(method='close',
                           kernel='1:0',
                           iterations='p')


def test_morphology_user_defined():
    with Image(filename='rose:') as img:
        was = img.signature
        img.morphology(method='dilate',
                       kernel='3x3: 0.3,0.6,0.3 0.6,1.0,0.6 0.3,0.6,0.3')
        assert was != img.signature
        was = img.signature
        img.morphology(method='dilate',
                       kernel='3x3: 0.3,0.6,0.3 0.6,1.0,0.6 0.3,0.6,0.3',
                       channel='red')
        assert was != img.signature
        with raises(ValueError):
            img.morphology(method='dilate',
                           kernel='junk:0')


def test_motion_blur():
    with Image(filename='rose:') as img:
        was = img.signature
        img.motion_blur(8, 6, 45)
        result = img.signature
        assert was != result
        was = result
        img.motion_blur(8, 6, -45, channel='blue')
        assert was != img.signature


def test_negate_default():
    with Image(width=100, height=100, pseudo='gradient:') as img:
        was = img.signature
        img.negate()
        assert was != img.signature
        assert img.negate(False, 'red')
        with raises(ValueError):
            img.negate(True, 'unknown')


def test_noise():
    with Image(filename='rose:') as img:
        was = img.signature
        img.noise('gaussian', 1.0)
        assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        img.noise('gaussian', 1.0, channel='red')
        assert was != img.signature


def test_normalize():
    with Image(width=100, height=100, pseudo='rose:') as img:
        was = img.signature
        img.normalize()
        assert was != img.signature
        with raises(ValueError):
            img.normalize('unknown')


def test_normalize_channel():
    with Image(width=100, height=100, pseudo='rose:') as img:
        was = img.signature
        img.normalize('red')
        assert was != img.signature


def test_oil_paint():
    with Image(filename='rose:') as img:
        was = img.signature
        img.oil_paint(5)
        assert was != img.signature


def test_opaque_paint():
    pink = Color('pink')
    white = Color('white')
    with Image(filename='WIZARD:') as img:
        img.opaque_paint(target=white, fill=pink,
                         fuzz=0.25*img.quantum_range)
        assert img[0, 0] == pink
    with Image(filename='WIZARD:') as img:
        img.opaque_paint(target='white', fill='pink',
                         fuzz=0.25*img.quantum_range, invert=True)
        assert img[0, 0] == white
    with Image(filename='WIZARD:') as img:
        was = img.signature
        img.opaque_paint(target='white', fill='pink',
                         fuzz=0.25*img.quantum_range, channel='red')
        assert was != img.signature


def test_optimize_layers(fx_asset):
    with Image(filename=str(fx_asset.joinpath('nocomments.gif'))) as img1:
        with Image(img1) as img2:
            img2.optimize_layers()
            assert img1.signature != img2.signature
            assert img1.size == img2.size


def test_optimize_transparency(fx_asset):
    with Image(filename=str(fx_asset.joinpath('nocomments.gif'))) as img1:
        with Image(img1) as img2:
            try:
                img2.optimize_transparency()
                assert img1.signature != img2.signature
                assert img1.size == img2.size
            except AttributeError as e:
                warnings.warn('MagickOptimizeImageTransparency not '
                              'present on system. ' + repr(e))


def test_ordered_dither():
    with Image(filename='rose:') as img:
        was = img.signature
        img.ordered_dither('o3x3')
        assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        img.ordered_dither('o3x3', channel='red')
        assert was != img.signature


def test_parse_meta_geometry():
    with Image(filename='rose:') as img:
        w, h, x, y = img.parse_meta_geometry('200%')
        assert (w, h, x, y) == (140, 92, 0, 0)
        with raises(ValueError):
            img.parse_meta_geometry('junk')


def test_percent_escape():
    with Image(filename='rose:') as img:
        assert '3019 70x46' == img.percent_escape('%k %wx%h')


def test_polaroid(fx_asset):
    # For testing polaroid method, we can't really identify if somethings
    # has changed correctly.
    with Image(filename='rose:') as img:
        img.polaroid()
    with Image(filename='rose:') as img:
        img.polaroid(caption='hello')
    with Image(filename='rose:') as img:
        font = Font(str(fx_asset.joinpath('League_Gothic.otf')), 12,
                    Color('orange'), True, Color('pink'), 1)
        img.polaroid(caption='hello', font=font)
        with raises(TypeError):
            img.polaroid(caption='hello', font='League_Gothic.otf')


@mark.skipif(MAGICK_VERSION_NUMBER < 0x708,
             reason='Polynomial requires ImageMagick-7.0.8.')
def test_polynomial():
    with Image(filename='rose:') as img:
        was = img.signature
        img.polynomial(arguments=(0.5, 1.0))
        assert was != img.signature


def test_posterize():
    with Image(filename='rose:') as img:
        was = img.signature
        img.posterize(levels=16, dither='no')
        assert was != img.signature
        with raises(TypeError):
            img.posterize(levels='16')
        with raises(ValueError):
            img.posterize(levels=16, dither='manhatten')


def test_quantize():
    number_colors = 32
    with Image(width=100, height=100, pseudo='gradient:') as img:
        assert img.colors > number_colors

    with Image(width=100, height=100, pseudo='gradient:') as img:
        with raises(TypeError):
            img.quantize(str(number_colors), 'undefined', 0, True, True)

        with raises(TypeError):
            img.quantize(number_colors, 0, 0, True, True)

        with raises(TypeError):
            img.quantize(number_colors, 'undefined', 'depth', True, True)

        with raises(TypeError):
            img.quantize(number_colors, 'undefined', 0, 1, True)

        with raises(TypeError):
            img.quantize(number_colors, 'undefined', 0, True, 1)

        img.quantize(number_colors, 'undefined', 0, True, True)
        assert img.colors <= number_colors


def test_random_threshold():
    with Image(filename='rose:') as img:
        was = img.signature
        img.random_threshold(low=0.4, high=0.6)
        assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        img.random_threshold(low=0.4, high=0.6, channel='red')
        assert was != img.signature


def test_range_channel():
    with Image(filename='rose:') as img:
        minima, maxima = img.range_channel('red')
        assert minima < maxima
        with raises(ValueError):
            img.range_channel('unknown')


@mark.skipif(MAGICK_VERSION_NUMBER < 0x708,
             reason='Range Threshold requires ImageMagick-7.0.8.')
def test_range_threshold():
    with Image(filename='rose:') as img:
        was = img.signature
        img.range_threshold(20, 40, 60, 80)
        assert was != img.signature
        # Smoke test
        img.range_threshold(20)
        img.range_threshold(20, 40)
        img.range_threshold(20, 40, 60)


def test_region():
    with Image(filename='rose:') as src:
        w, h, x, y = src.page
        with src.region() as dst:
            assert (w, h) == dst.size
        with src.region(width=w//2, height=h//2, x=x//2, y=y//2) as dst:
            assert (w//2, h//2) == dst.size
    with Image(filename='rose:') as src:
        with src.region(width=10, height=10, gravity='south_east') as dst:
            assert (70, 46, 60, 36) == dst.page
            assert (10, 10) == dst.size
    with Image(filename='rose:') as img:
        with img.region(x=10, gravity='center') as dst:
            assert (70, 46, 10, 0) == dst.page
            assert (60, 46) == dst.size


def test_remap():
    with Image(filename='rose:') as img:
        was = img.signature
        with Image(filename='hald:3') as palette:
            img.remap(palette)
        assert was != img.signature
        with raises(TypeError):
            img.remap(0xDEADBEEF)
        with raises(TypeError):
            img.remap(img, method=0xDEADBEEF)
        with raises(ValueError):
            img.remap(img, method='none')


@mark.parametrize(('density', 'expected_size'), [
    ((72, 72), (800, 600)),
    ((36, 36), (400, 300)),
    ((144, 144), (1600, 1200)),
    ((None, 36), (800, 300)),
    ((36, None), (400, 600)),
])
def test_resample(density, expected_size, fx_asset):
    """Resample (Adjust number of pixels at the given density) the image."""
    xr, yr = density
    with Image(filename=str(fx_asset.joinpath('beach.jpg'))) as img:
        img.units = "pixelspercentimeter"
        assert img.resolution == (72, 72)
        img.resample(xr, yr)
        # Expect ``None`` values to match ImageMagick's default 72 resolution.
        if xr is None:
            xr = 72
        if yr is None:
            yr = 72
        assert img.resolution == (xr, yr)
        assert img.size == expected_size


def test_resample_errors():
    """Sampling errors."""
    with Image(filename='rose:') as img:
        with raises(TypeError):
            img.resample(x_res='100')
        with raises(TypeError):
            img.resample(y_res='100')
        with raises(ValueError):
            img.resample(x_res=0)
        with raises(ValueError):
            img.resample(y_res=0)
        with raises(ValueError):
            img.resample(x_res=-5)
        with raises(ValueError):
            img.resample(y_res=-5)


@mark.parametrize(('method'), [
    ('resize'),
    ('sample'),
])
def test_resize_and_sample(method):
    """Resizes/Samples the image."""
    with Image(filename='wizard:') as img:
        with img.clone() as a:
            assert a.size == (480, 640)
            getattr(a, method)(100, 100)
            assert a.size == (100, 100)
        with img.clone() as b:
            assert b.size == (480, 640)
            getattr(b, method)(height=100)
            assert b.size == (480, 100)
        with img.clone() as c:
            assert c.size == (480, 640)
            getattr(c, method)(width=100)
            assert c.size == (100, 640)


@mark.parametrize(('method'), [
    ('resize'),
    ('sample'),
])
def test_resize_and_sample_errors(method):
    """Resizing/Sampling errors."""
    with Image(filename='rose:') as img:
        with raises(TypeError):
            getattr(img, method)(width='100')
        with raises(TypeError):
            getattr(img, method)(height='100')
        with raises(ValueError):
            getattr(img, method)(width=0)
        with raises(ValueError):
            getattr(img, method)(height=0)
        with raises(ValueError):
            getattr(img, method)(width=-5)
        with raises(ValueError):
            getattr(img, method)(height=-5)


@mark.slow
@mark.parametrize(('method'), [
    ('resize'),
    ('sample'),
])
def test_resize_and_sample_gif(method, tmp_path, fx_asset):
    fpath = str(fx_asset.joinpath('nocomments-delay-100.gif'))
    with Image(filename=fpath) as img:
        assert len(img.sequence) == 46
        with img.clone() as a:
            assert a.size == (350, 197)
            assert a.sequence[0].delay == 100
            for s in a.sequence:
                assert s.delay == 100
            getattr(a, method)(175, 98)
            a.save(filename=str(tmp_path / '175_98.gif'))
        with Image(filename=str(tmp_path / '175_98.gif')) as a:
            assert len(a.sequence) == 46
            assert a.size == (175, 98)
            for s in a.sequence:
                assert s.delay == 100
        with img.clone() as b:
            assert b.size == (350, 197)
            for s in b.sequence:
                assert s.delay == 100
            getattr(b, method)(height=100)
            b.save(filename=str(tmp_path / '350_100.gif'))
        with Image(filename=str(tmp_path / '350_100.gif')) as b:
            assert len(b.sequence) == 46
            assert b.size == (350, 100)
            for s in b.sequence:
                assert s.delay == 100
        with img.clone() as c:
            assert c.size == (350, 197)
            for s in c.sequence:
                assert s.delay == 100
            getattr(c, method)(width=100)
            c.save(filename=str(tmp_path / '100_197.gif'))
        with Image(filename=str(tmp_path / '100_197.gif')) as c:
            assert len(c.sequence) == 46
            assert c.size == (100, 197)
            for s in c.sequence:
                assert s.delay == 100


def test_roll():
    with Image(filename='rose:') as img:
        was = img.signature
        img.roll(x=-15, y=15)
        assert was != img.signature


@mark.slow
def test_rotate(fx_asset):
    """Rotates an image."""
    with Image(filename=str(fx_asset.joinpath('rotatetest.gif'))) as img:
        assert 150 == img.width
        assert 100 == img.height
        with img.clone() as cloned:
            cloned.rotate(360)
            assert img.size == cloned.size
            with Color('black') as black:
                assert black == cloned[0, 50] == cloned[74, 50]
                assert black == cloned[0, 99] == cloned[74, 99]
            with Color('white') as white:
                assert white == cloned[75, 50] == cloned[75, 99]
        with img.clone() as cloned:
            cloned.rotate(90)
            assert 100 == cloned.width
            assert 150 == cloned.height
            with Color('black') as black:
                with Color('white') as white:
                    for y, row in enumerate(cloned):
                        for x, col in enumerate(row):
                            if y < 75 and x < 50:
                                assert col == black
                            else:
                                assert col == white
        with Color('red') as bg:
            with img.clone() as cloned:
                cloned.rotate(45, bg)
                assert 176 <= cloned.width == cloned.height <= 178
                assert bg == cloned[0, 0] == cloned[0, -1]
                assert bg == cloned[-1, 0] == cloned[-1, -1]
                with Color('black') as black:
                    # Until we implement antialiasing, we need to evaluate
                    # pixels next to corners.
                    assert black == cloned[5, 70]
                    assert black == cloned[36, 39]
                    assert black == cloned[85, 88]
                    assert black == cloned[53, 120]
        with Color('red') as bg:
            with img.clone() as cloned:
                cloned.rotate(45, 'red')
                assert 176 <= cloned.width == cloned.height <= 178
                assert bg == cloned[0, 0] == cloned[0, -1]
                assert bg == cloned[-1, 0] == cloned[-1, -1]
                with Color('black') as black:
                    # Until we implement antialiasing, we need to evaluate
                    # pixels next to corners.
                    assert black == cloned[5, 70]
                    assert black == cloned[36, 39]
                    assert black == cloned[85, 88]
                    assert black == cloned[53, 120]


@mark.slow
def test_rotate_gif(tmp_path, fx_asset):
    fpath = str(fx_asset.joinpath('nocomments-delay-100.gif'))
    with Image(filename=fpath) as img:
        for s in img.sequence:
            assert s.delay == 100
        with img.clone() as e:
            assert e.size == (350, 197)
            e.rotate(90)
            for s in e.sequence:
                assert s.delay == 100
            e.save(filename=str(tmp_path / 'rotate_90.gif'))
        with Image(filename=str(tmp_path / 'rotate_90.gif')) as e:
            assert e.size == (197, 350)
            assert len(e.sequence) == 46
            for s in e.sequence:
                assert s.delay == 100


def test_rotate_reset_coords():
    """Reset the coordinate frame so to the upper-left corner of
    the image is (0, 0) again.

    """
    with Image(filename='rose:') as img:
        img.rotate(45, reset_coords=True)
        img.crop(0, 0, 84, 84)
        # There should be no page info from the crop, as everything was
        # nulled by the rotate.
        assert (0, 0, 0, 0) == img.page


@mark.skipif(MAGICK_VERSION_NUMBER < 0x688,
             reason="Not supported until after ImageMagick-6.8.8")
def test_rotational_blur():
    with Image(filename='rose:') as img:
        was = img.signature
        img.rotational_blur(45.0)
        now = img.signature
        assert was != now
        was = now
        img.rotational_blur(180, 'blue')
        assert was != img.signature


def test_scale():
    with Image(filename='rose:') as img:
        width, height = img.size
        img.scale(2, 3)
        assert width*2, height*3 == img.size


def test_selective_blur():
    with Image(filename='rose:') as img:
        was = img.signature
        img.selective_blur(8, 3, 0.1 * img.quantum_range)
        assert was != img.signature
        was = img.signature
        img.selective_blur(8, 3, 0.1 * img.quantum_range, channel='red')
        assert was != img.signature


def test_sepia_tone():
    with Image(filename='rose:') as img:
        was = img.signature
        img.sepia_tone(threshold=0.8)
        assert was != img.signature


def test_shade():
    with Image(filename='rose:') as img:
        was = img.signature
        img.shade(gray=False, azimuth=10.0, elevation=10.0)
        assert was != img.signature
        with raises(TypeError):
            img.shade(azimuth='hello')
        with raises(TypeError):
            img.shade(elevation='hello')


def test_shadow():
    with Image(filename='rose:') as img:
        was = img.size
        img.shadow(alpha=5.0, sigma=1.25, x=10, y=10)
        assert was != img.size
        with raises(TypeError):
            img.shadow(alpha='hello')
        with raises(TypeError):
            img.shadow(sigma='hello')
        with raises(TypeError):
            img.shadow(x=None)
        with raises(TypeError):
            img.shadow(y=None)


def test_sharpen():
    with Image(filename='rose:') as img:
        was = img.signature
        img.sharpen(radius=10.0, sigma=2.0)
        assert was != img.signature
        with raises(TypeError):
            img.sharpen(radius='hello')
        with raises(TypeError):
            img.sharpen(sigma='hello')
    with Image(filename='rose:') as img:
        was = img.signature
        img.sharpen(radius=10.0, sigma=2.0, channel='red')
        assert was != img.signature


def test_shave():
    with Image(filename='rose:') as img:
        was = img.size
        img.shave(10, 10)
        assert was != img.size
        with raises(TypeError):
            img.shave(None, 10)
        with raises(TypeError):
            img.shave(10, None)


def test_shear():
    with Image(filename='rose:') as img:
        was = img.signature
        img.shear(background='green', x=10, y=10)
        assert was != img.signature


def test_sigmoidal_contrast():
    with Image(filename='rose:') as img:
        was = img.signature
        img.sigmoidal_contrast(sharpen=True,
                               strength=3,
                               midpoint=0.65 * img.quantum_range)
        assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        img.sigmoidal_contrast(sharpen=True,
                               strength=3,
                               midpoint=0.65 * img.quantum_range,
                               channel='red')
        assert was != img.signature


def test_similarity():
    with Image(filename='rose:') as img:
        with img.clone() as sub_img:
            location, diff = img.similarity(sub_img)
            assert location['top'] == 0 and location['left'] == 0
            assert diff < 0.001
        with raises(TypeError):
            img.similarity(0xDEADBEEF)


def test_sketch():
    with Image(filename='rose:') as img:
        was = img.signature
        img.sketch(5.0, 3.0, 45.0)
        assert was != img.signature


def test_smush():
    with Image(filename='rose:') as img:
        width, height = img.size
        with img.clone() as a:
            with img.clone() as b:
                a.sequence.append(b)
                a.smush(False, 10)
                assert a.size == (width * 2 + 10, height)
        with img.clone() as a:
            with img.clone() as b:
                a.sequence.append(b)
                a.smush(True, 0)
                assert a.size == (width, height * 2)
        with raises(TypeError):
            img.smush(0xDEADBEEF, '0x0')


def test_solarize():
    with Image(filename='rose:') as img:
        was = img.signature
        img.alpha_channel = 'off'  # Needed for IM-7
        img.solarize(0.5 * img.quantum_range)
        assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        img.alpha_channel = 'off'  # Needed for IM-7
        img.solarize(0.5 * img.quantum_range, channel='red')
        assert was != img.signature


def test_sparse_color():
    with Image(width=10, height=10, background='white') as img:
        was = img.signature
        colors = [
            (0, 0, '#F00'),
            (9, 9, '#00F'),
        ]
        img.sparse_color('barycentric', colors)
        assert was != img.signature
        with raises(TypeError):
            img.sparse_color(0xDEADBEEF, colors)
        with raises(TypeError):
            img.sparse_color('barycentric', 0xDEADBEEF)
        with raises(TypeError):
            img.sparse_color('barycentric', colors, channel_mask='red')


def test_splice():
    green = Color('GREEN')
    with Image(filename='rose:') as img:
        width, height = img.size
        img.background_color = green
        img.splice(10, 10, 10, 10)
        assert width+10, height+10 == img.size
        assert img[15, 15] == green
    with Image(filename='rose:') as img:
        was = img.signature
        img.splice(width=10, height=10, gravity='center')
        assert img.signature != was
    with Image(filename='rose:') as img:
        img.splice(width=10, height=10, x=10, gravity='center')
        assert (80, 56) == img.size


def test_spread():
    with Image(filename='rose:') as img:
        was = img.signature
        img.spread(8.0)
        assert was != img.signature


def test_statistic():
    with Image(filename='rose:') as img:
        was = img.signature
        img.statistic('median', 5, 5)
        assert was != img.signature
    with Image(filename='rose:') as img:
        was = img.signature
        img.statistic('median', 5, 5, channel='red')
        assert was != img.signature


def test_stegano():
    with Image(filename='wizard:') as img:
        was = img.signature
        with Image(filename='rose:') as watermark:
            img.stegano(watermark)
        assert was != img.signature
        with raises(TypeError):
            img.stegano(0xDEADBEEF)


def test_strip(fx_asset):
    """Strips the image of all profiles and comments."""
    with Image(filename=str(fx_asset.joinpath('beach.jpg'))) as img:
        strio = io.BytesIO()
        img.save(file=strio)
        len_original = strio.tell()
        strio.close()
        strio = io.BytesIO()
        img.strip()
        img.save(file=strio)
        len_stripped = strio.tell()
        assert len_original > len_stripped


def test_swirl():
    with Image(filename='rose:') as img:
        was = img.signature
        img.swirl(degree=90)
        assert was != img.signature


def test_texture():
    with Image(filename='rose:') as img:
        was = img.signature
        with Image(width=1, height=10, pseudo='gradient:') as tile:
            img.texture(tile)
        assert was != img.signature
        with raises(TypeError):
            img.texture(0xDEADBEEF)


def test_threshold():
    with Image(width=100, height=100, pseudo='gradient:white-black') as img:
        top = int(img.height * 0.25)
        btm = int(img.height * 0.75)
        img.threshold(0.5)
        with img[0, top] as white:
            assert white.red_int8 == white.green_int8 == white.blue_int8 == 255
        with img[0, btm] as black:
            assert black.red_int8 == black.green_int8 == black.blue_int8 == 0
        with raises(ValueError):
            img.threshold(0.5, channel='unknown')


def test_threshold_channel():
    with Image(width=100, height=100, pseudo='gradient:white-black') as img:
        was = img.signature
        img.threshold(0.0, 'red')
        img.threshold(0.5, 'green')
        img.threshold(1.0, 'blue')
        assert was != img.signature


def test_thumbnail():
    with Image(filename='rose:') as img:
        img.thumbnail(50, 20)
        assert (50, 20) == img.size
    with Image(filename='rose:') as img:
        img.thumbnail(width=50)
        assert (50, img.height) == img.size
    with Image(filename='rose:') as img:
        img.thumbnail(height=20)
        assert (img.width, 20) == img.size


def test_tint():
    with Image(filename='rose:') as img:
        was = img.signature
        img.tint('blue', 'blue')
        assert was != img.signature
        with raises(TypeError):
            img.colorize(0xDEADBEEF, Color('blue'))
        with raises(TypeError):
            img.colorize(Color('blue'), 0xDEADBEEF)


@mark.parametrize(('args', 'kwargs', 'expected_size'), [
    ((), {'resize': '200%'}, (960, 1280)),
    ((), {'resize': '200%x100%'}, (960, 640)),
    ((), {'resize': '1200'}, (1200, 1600)),
    ((), {'resize': 'x300'}, (225, 300)),
    ((), {'resize': '400x600'}, (400, 533)),
    ((), {'resize': '1000x1200^'}, (1000, 1333)),
    ((), {'resize': '100x100!'}, (100, 100)),
    ((), {'resize': '400x500>'}, (375, 500)),
    ((), {'resize': '1200x3000<'}, (1200, 1600)),
    ((), {'resize': '120000@'}, (300, 400)),
    ((), {'crop': '300x300'}, (300, 300)),
    ((), {'crop': '300x300+100+100'}, (300, 300)),
    ((), {'crop': '300x300-150-150'}, (150, 150)),
    (('300x300', '200%'), {}, (600, 600)),
])
def test_transform(args, kwargs, expected_size):
    """Transforms (crops and resizes with geometry strings) the image."""
    with Image(filename='wizard:') as img:
        assert img.size == (480, 640)
        img.transform(*args, **kwargs)
        assert img.size == expected_size


@mark.skipif(MAGICK_VERSION_NUMBER < 0x708,
             reason="Crop by aspect-ration requires ImageMagick-7.0.8")
def test_transform_aspect_crop():
    with Image(filename='wizard:') as img:
        img.transform(crop='16:9')
        assert img.size == (480, 270)


def test_transform_colorspace():
    with Image(filename='hald:3') as img:
        with raises(TypeError):
            img.transform_colorspace(0xDEADBEEF)
        with raises(ValueError):
            img.transform_colorspace('unknown')

        img.transform_colorspace('srgb')
        assert img.colorspace == 'srgb'


def test_transform_errors():
    """Tests errors raised by invalid parameters for transform."""
    unichar = b'\xe2\x9a\xa0'.decode('utf-8')
    with Image(filename='wizard:') as img:
        with raises(TypeError):
            img.transform(crop=500)
        with raises(TypeError):
            img.transform(resize=500)
        with raises(TypeError):
            img.transform(500, 500)
        with raises(ValueError):
            img.transform(crop=unichar)
        with raises(ValueError):
            img.transform(resize=unichar)


def test_transform_gif(tmp_path, fx_asset):
    src = str(fx_asset / 'nocomments-delay-100.gif')
    dst = str(tmp_path / 'test_transform_gif.gif')
    with Image(filename=src) as img:
        assert len(img.sequence) == 46
        assert img.size == (350, 197)
        for single in img.sequence:
            assert single.delay == 100
        img.transform(resize='175x98!')
        assert len(img.sequence) == 46
        assert img.size == (175, 98)
        for single in img.sequence:
            assert single.size == (175, 98)
            assert single.delay == 100
        img.save(filename=dst)
    with Image(filename=dst) as gif:
        assert len(gif.sequence) == 46
        assert gif.size == (175, 98)
        for single in gif.sequence:
            assert single.size == (175, 98)
            assert single.delay == 100


def test_transparent_color(fx_asset):
    """TransparentPaint test
    .. versionchanged:: 0.5.0
       Alpha channel must be enabled with ``'set'``, previously ``True``.
       See docstring in :meth:`wand.image.BaseImage.alpha_channel`.
    """
    with Image(filename=str(fx_asset.joinpath('rotatetest.gif'))) as img:
        img.alpha_channel = 'set'
        with Color('white') as white:
            img.transparent_color(white, 0.0, 2, 0)
            assert img[75, 50].alpha == 0
            assert img[0, 50].alpha == 1.0
    with Image(filename=str(fx_asset.joinpath('rotatetest.gif'))) as img:
        img.alpha_channel = 'set'
        img.transparent_color('white', 0.0, 2, 0)
        assert img[75, 50].alpha == 0
        assert img[0, 50].alpha == 1.0


def test_transparentize(fx_asset):
    with Image(filename=str(fx_asset.joinpath('croptest.png'))) as im:
        with Color('transparent') as transparent:
            with Color('black') as black:
                assert im[99, 100] == transparent
                assert im[100, 100] == black
                im.transparentize(0.3)
                im.clamp()
                assert im[99, 100].alpha_int8 == transparent.alpha_int8
                with im[100, 100] as c:
                    assert c.red == c.green == c.blue == 0
                    assert 0.69 < c.alpha < 0.71
        with raises(ValueError):
            im.transparentize(-9)


def test_transpose():
    with Image(filename='rose:') as img:
        was = img.signature
        img.transpose()
        assert was != img.signature


def test_transverse():
    with Image(filename='rose:') as img:
        was = img.signature
        img.transverse()
        assert was != img.signature


def test_trim():
    """Remove transparent area around image."""
    with Image(filename='logo:') as img:
        old_x, _ = img.size
        img.trim()
        new_x, _ = img.size
        assert new_x < old_x


def test_trim_color():
    with Image(filename='wizard:') as img:
        size = img.size
        img.trim(color='white', fuzz=0.1*img.quantum_range)
        assert img.size != size


@mark.skipif(MAGICK_VERSION_NUMBER < 0x709,
             reason='Trim by percent-background requires ImagesMagick-7.0.9')
def test_trim_percent_background():
    with Image(filename='wizard:') as img:
        was = img.size
        img.trim(fuzz=0.0, percent_background=0.5, background_color='white')
        assert img.size != was


def test_trim_reset_coords():
    with Image(filename='logo:') as img:
        page = img.page
        img.trim(reset_coords=True)
        assert page != img.page


def test_unique_colors():
    with Image(filename='rose:') as img:
        was = img.signature
        img.unique_colors()
        assert was != img.signature


def test_unsharp_mask():
    with Image(filename='rose:') as img:
        was = img.signature
        alpha = was
        with img.clone() as cpy:
            cpy.unsharp_mask(1.1, 1, 0.5, 0.001)
            alpha = cpy.signature
            assert alpha != was
        with img.clone() as cpy:
            cpy.unsharp_mask(1.1, 1, 0.5, 0.001, channel='red')
            assert cpy.signature != alpha


def test_vignette():
    with Image(filename='rose:') as img:
        was = img.signature
        img.vignette(radius=3, sigma=3)
        assert was != img.signature


def test_watermark():
    """Adds watermark to an image."""
    with Image(filename='wizard:') as img:
        was = img.signature
        with Image(filename='rose:') as wm:
            img.watermark(wm, 0.3)
            assert was != img.signature


def test_wave():
    with Image(filename='rose:') as img:
        was = img.size
        img.wave(amplitude=img.height, wave_length=img.width/2)
        assert was != img.size
        with raises(TypeError):
            img.wave(amplitude='img height')
        with raises(TypeError):
            img.wave(wave_length='img height')
        with raises(TypeError):
            img.wave(method=0xDEADBEEF)


@mark.skipif(MAGICK_VERSION_NUMBER < 0x708,
             reason='Wavelet Denoise requires ImageMagick-7.0.8.')
def test_wavelet_denoise():
    with Image(filename='rose:') as img:
        was = img.signature
        img.wavelet_denoise(0.2, 0.3)
        assert was != img.signature


@mark.skipif(MAGICK_VERSION_NUMBER < 0x70B,
             reason='White balance requires ImageMagick-7.0.11')
def test_white_balance():
    with Image(filename='rose:') as img:
        was = img.signature
        img.white_balance()
        assert was != img.signature


def test_white_threshold():
    with Image(filename='rose:') as img:
        was = img.signature
        img.white_threshold(Color('gray(50%)'))
        assert was != img.signature
        assert img.white_threshold('gray(50%)')
        with raises(TypeError):
            img.white_threshold(0xDEADBEEF)



from pytest import mark, raises

from wand.api import library
from wand.image import Image


def expire(image):
    """Expire image's sequence cache."""
    image.sequence.instances = []


def test_length(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as img:
        assert len(img.sequence) == 4


def test_validate_position_error(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as img:
        with raises(TypeError):
            img.sequence.validate_position(0.001)


def test_validate_slice_error(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as img:
        with raises(ValueError):
            img.sequence.validate_slice(slice(0, 10, 3))


def test_getitem(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as img:
        size = img.size
        assert size == img.sequence[img.sequence.current_index].size
        assert img.sequence[0].size == (32, 32)
        assert img.sequence[1].size == (16, 16)
        assert img.sequence[2].size == (32, 32)
        assert img.sequence[3].size == (16, 16)
        with raises(IndexError):
            img.sequence[4]
        assert img.sequence[-1].size == (16, 16)
        assert img.sequence[-2].size == (32, 32)
        assert img.sequence[-3].size == (16, 16)
        assert img.sequence[-4].size == (32, 32)
        with raises(IndexError):
            img.sequence[-5]
        assert img.size == size


def test_setitem(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as imga:
        with Image(filename=str(fx_asset.joinpath('google.ico'))) as imgg:
            imga.sequence[2] = imgg
        assert len(imga.sequence) == 4
        assert imga.sequence[2].size == (16, 16)
        expire(imga)
        assert imga.sequence[2].size == (16, 16)
        with raises(TypeError):
            imga.sequence[3] = 0xDEADBEEF


def test_delitem(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as img:
        detached = img.sequence[0]
        del img.sequence[0]
        assert len(img.sequence) == 3
        assert img.sequence[0] is not detached
        assert img.sequence[0].size == (16, 16)
        expire(img)
        assert len(img.sequence) == 3
        assert img.sequence[0] is not detached
        assert img.sequence[0].size == (16, 16)
        assert img.sequence[1].size == (32, 32)
        assert img.sequence[2].size == (16, 16)


def test_delitem_not_loaded(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as img:
        del img.sequence[1]
        assert len(img.sequence) == 3


slices = {
    'to_end': slice(2, None, None),
    'from_first': slice(None, 2, None),
    'from_back': slice(-2, None, None),
    'to_back': slice(None, -2, None),
    'middle': slice(1, 3, None),
    'from_overflow': slice(10, None, None),
    'to_overflow': slice(None, 10, None)
}

slice_items = list(sorted(slices.items(), key=lambda k: k[0]))


@mark.parametrize(('slice_name', 'slice_'), slice_items)
def test_getitem_slice(slice_name, slice_, fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as img:
        assert list(img.sequence[slice_]) == list(img.sequence)[slice_]


@mark.parametrize(('slice_name', 'slice_'), slice_items)
def test_setitem_slice(slice_name, slice_, fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as imga:
        instances = list(imga.sequence)
        with Image(filename=str(fx_asset.joinpath('github.ico'))) as imgg:
            instances[slice_] = imgg.sequence
            imga.sequence[slice_] = imgg.sequence
            assert instances == list(imga.sequence)
            expire(imga)
            assert instances == list(imga.sequence)


@mark.parametrize(('slice_name', 'slice_'), slice_items)
def test_delitem_slice(slice_name, slice_, fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as img:
        instances = list(img.sequence)
        del instances[slice_]
        del img.sequence[slice_]
        assert list(img.sequence) == instances
        expire(img)
        assert list(img.sequence) == instances


def test_iterator(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as img:
        container_size = img.sequence[img.sequence.current_index].size
        actual = []
        expected = [(32, 32), (16, 16), (32, 32), (16, 16)]
        for i in img.sequence:
            actual.append(i.size)
            assert img.size == container_size
        assert actual == expected


def test_append(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as imga:
        with Image(filename=str(fx_asset.joinpath('google.ico'))) as imgg:
            imga.sequence.append(imgg)
            assert imga.sequence[4] == imgg.sequence[0]
            expire(imga)
            assert imga.sequence[4] == imgg.sequence[0]
        assert len(imga.sequence) == 5
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as imga:
        with Image(filename=str(fx_asset.joinpath('github.ico'))) as imgg:
            imga.sequence.append(imgg)
            assert imga.sequence[4] == imgg.sequence[0]
            expire(imga)
            assert imga.sequence[4] == imgg.sequence[0]
        assert len(imga.sequence) == 5
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as imga:
        with Image(filename=str(fx_asset.joinpath('github.ico'))) as imgg:
            imga.sequence.append(imgg.sequence[1])
            assert imga.sequence[4] == imgg.sequence[1]
            expire(imga)
            assert imga.sequence[4] == imgg.sequence[1]
        assert len(imga.sequence) == 5


def test_insert(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as imga:
        instances = [imga.sequence[i] for i in range(2, 4)]
        assert len(imga.sequence) == 4
        with Image(filename=str(fx_asset.joinpath('github.ico'))) as imgg:
            imga.sequence.insert(2, imgg)
            assert imga.sequence[2] == imgg.sequence[0]
            assert len(imga.sequence) == 5
            for i, instance in enumerate(instances):
                assert instance == imga.sequence[3 + i]
            expire(imga)
            assert imga.sequence[2] == imgg.sequence[0]
        for i, instance in enumerate(instances):
            assert instance == imga.sequence[3 + i]
        with raises(TypeError):
            imga.sequence.insert(0xDEADBEEF)


def test_insert_first(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as imga:
        assert len(imga.sequence) == 4
        with Image(filename=str(fx_asset.joinpath('github.ico'))) as imgg:
            imga.sequence.insert(0, imgg)
            assert imga.sequence[0] == imgg.sequence[0]
            expire(imga)
            assert imga.sequence[0] == imgg.sequence[0]
        assert len(imga.sequence) == 5


def test_extend(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as a:
        length = len(a.sequence)
        with Image(filename=str(fx_asset.joinpath('github.ico'))) as b:
            a.sequence.extend(list(b.sequence)[::-1])
            assert a.sequence[length] == b.sequence[1]
            assert a.sequence[length + 1] == b.sequence[0]
            expire(a)
            assert a.sequence[length] == b.sequence[1]
            assert a.sequence[length + 1] == b.sequence[0]
        assert len(a.sequence) == 6
        with raises(TypeError):
            a.sequence.extend([0xDEADBEEF])


def test_extend_sequence(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as a:
        length = len(a.sequence)
        with Image(filename=str(fx_asset.joinpath('github.ico'))) as b:
            a.sequence.extend(b.sequence)
            for i in range(2):
                assert a.sequence[length + i] == b.sequence[i]
            expire(a)
            for i in range(2):
                assert a.sequence[length + i] == b.sequence[i]
        assert len(a.sequence) == 6


@mark.parametrize('how_many', range(2, 5))
def test_extend_offset(fx_asset, how_many):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as a:
        instances = list(a.sequence)
        with Image(filename=str(fx_asset.joinpath('github.ico'))) as b:
            added = list(b.sequence)[::-1] + [b.sequence[0]] * (how_many - 2)
            a.sequence.extend(added, 2)
            instances[2:2] = added
            assert list(a.sequence) == instances
            expire(a)
            assert list(a.sequence) == instances
        assert len(a.sequence) == 4 + how_many


def test_extend_offset_sequence(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as a:
        instances = list(a.sequence)
        with Image(filename=str(fx_asset.joinpath('github.ico'))) as b:
            a.sequence.extend(b.sequence, 2)
            instances[2:2] = list(b.sequence)
            assert list(a.sequence) == instances
            expire(a)
            assert list(a.sequence) == instances
        assert len(a.sequence) == 6


def test_extend_first(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as a:
        instances = list(a.sequence)
        with Image(filename=str(fx_asset.joinpath('github.ico'))) as b:
            a.sequence.extend(list(b.sequence)[::-1], 0)
            instances[:0] = list(b.sequence)[::-1]
            assert list(a.sequence) == instances
            expire(a)
            assert list(a.sequence) == instances
        assert len(a.sequence) == 6


def test_extend_first_sequence(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as a:
        instances = list(a.sequence)
        with Image(filename=str(fx_asset.joinpath('github.ico'))) as b:
            a.sequence.extend(b.sequence, 0)
            instances[:0] = list(b.sequence)
            assert list(a.sequence) == instances
            expire(a)
            assert list(a.sequence) == instances
        assert len(a.sequence) == 6


def test_container(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as img:
        for single in img.sequence:
            assert single.container is img


def test_index(fx_asset):
    with Image(filename=str(fx_asset.joinpath('nocomments.gif'))) as img:
        for i, single in enumerate(img.sequence):
            assert single.index == i
        del img.sequence[0]
        for i, single in enumerate(img.sequence):
            assert single.index == i
    with Image(filename=str(fx_asset.joinpath('nocomments.gif'))) as img:
        del img.sequence[2]
        for i, single in enumerate(img.sequence):
            assert single.index == i


@mark.parametrize(('cmp_name', 'f'), [
    ('equals', lambda i: i),  # identity
    ('hash_equals', hash),
    ('signature_equals', lambda i: i.signature)
])
def test_equals(cmp_name, f, fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as a:
        with Image(filename=str(fx_asset.joinpath('apple.ico'))) as b:
            assert f(a) == f(b)
            assert f(a.sequence[0]) == f(b.sequence[0])
            assert f(a) != f(b.sequence[1])
            assert f(a.sequence[0]) != f(b.sequence[1])
            assert f(a.sequence[1]) == f(b.sequence[1])


def test_clone(fx_asset):
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as img:
        with img.sequence[2].clone() as single:
            assert single.wand != img.wand
            assert len(single.sequence) == 1
            assert len(list(single.sequence)) == 1
            assert single.size == img.sequence[2].size


def test_changes_reflected_back(fx_asset):
    """Changes on each single image should be reflected back to
    the container image.

    """
    with Image(filename=str(fx_asset.joinpath('apple.ico'))) as img:
        with img.sequence[3] as single:
            single.resize(32, 32)
            assert single.size == (32, 32)
            img.sequence.instances[3] = None  # to make new instance
            uncommitted = img.sequence[3]
            assert uncommitted.size == (16, 16)
        img.sequence.instances[3] = None
        with img.sequence[3] as committed:
            assert committed.size == (32, 32)


def test_delay(fx_asset):
    fpath = str(fx_asset.joinpath('nocomments-delay-100.gif'))
    with Image(filename=fpath) as img:
        for s in img.sequence:
            assert s.delay == 100


def test_set_delay(fx_asset):
    with Image(filename=str(fx_asset.joinpath('nocomments.gif'))) as img:
        with img.sequence[2] as frame:
            assert frame.delay == 0
            frame.delay = 10
        with img.sequence.index_context(2):
            assert library.MagickGetImageDelay(img.wand) == 10
        with raises(TypeError):
            img.sequence[2].delay = 0.0001
        with raises(ValueError):
            img.sequence[2].delay = -1



from pathlib import Path
from pytest import mark
from wand.version import VERSION
try:
    import tomllib
except ImportError:
    tomllib = None

CFG_ROOT = Path(__file__).parent.parent
CFG_PATH = CFG_ROOT / "pyproject.toml"

def test_pyproject_exists():
    assert CFG_PATH.exists()

@mark.skipif(tomllib is None, reason="tomllib not available")
def test_pyproject_valid():
    with open(CFG_PATH, 'rb') as fd:
        cfg = tomllib.load(fd)
    assert 'project' in cfg
    project = cfg.get('project', {})
    assert project.get('name') in ['wand', 'Wand']
    assert project.get('version') == VERSION


import itertools

from pytest import mark, raises

from wand.api import library
from wand.color import Color
from wand.drawing import Drawing
from wand.image import Image
from wand.version import MAGICK_VERSION_NUMBER


def test_init_user_error():
    with raises(TypeError):
        with Drawing(0xDEADBEEF):
            pass


def test_is_drawing_wand():
    with Drawing() as ctx:
        assert library.IsDrawingWand(ctx.resource)


def test_set_get_border_color():
    with Drawing() as ctx:
        with Color("#0F0") as green:
            ctx.border_color = green
            assert green == ctx.border_color
        ctx.border_color = 'orange'
        assert ctx.border_color == Color('orange')
        # Assert user error
        with raises(TypeError):
            ctx.border_color = 0xDEADBEEF


def test_set_get_clip_path():
    with Drawing() as ctx:
        ctx.clip_path = 'path_id'
        assert ctx.clip_path == 'path_id'
        # Assert user error
        with raises(TypeError):
            ctx.clip_path = 0xDEADBEEF


def test_set_get_clip_rule():
    with Drawing() as ctx:
        ctx.clip_rule = 'evenodd'
        assert ctx.clip_rule == 'evenodd'
        with raises(TypeError):
            ctx.clip_rule = 0xDEADBEEF
        with raises(ValueError):
            ctx.clip_rule = 'not-a-rule'


def test_set_get_clip_units():
    with Drawing() as ctx:
        ctx.clip_units = 'object_bounding_box'
        assert ctx.clip_units == 'object_bounding_box'
        with raises(TypeError):
            ctx.clip_units = 0xDEADBEEF
        with raises(ValueError):
            ctx.clip_units = 'not-a-clip_unit'


def test_set_get_font():
    with Drawing() as ctx:
        """Setting this values doesn't actually check if the typeface file
        exists, but the get/set values should still agree."""
        ctx.font = 'GhostType.ttf'
        assert ctx.font == 'GhostType.ttf'
        with raises(TypeError):
            ctx.font = 0xDEADBEEF


def test_set_get_font_family():
    with Drawing() as ctx:
        assert ctx.font_family is None
        ctx.font_family = 'sans-serif'
        assert ctx.font_family == 'sans-serif'
        with raises(TypeError):
            ctx.font_family = 0xDEADBEEF


def test_set_get_font_resolution():
    with Drawing() as ctx:
        ctx.font_resolution = (78.0, 78.0)
        assert ctx.font_resolution == (78.0, 78.0)
        with raises(TypeError):
            ctx.font_resolution = 0xDEADBEEF
        with raises(ValueError):
            ctx.font_resolution = (78.0, 78.0, 78.0)


def test_set_get_font_size():
    with Drawing() as ctx:
        ctx.font_size = 22.2
        assert ctx.font_size == 22.2
        with raises(TypeError):
            ctx.font_size = '22.2%'
        with raises(ValueError):
            ctx.font_size = -22.2


def test_set_get_font_stretch():
    with Drawing() as ctx:
        ctx.font_stretch = 'condensed'
        assert ctx.font_stretch == 'condensed'
        with raises(TypeError):
            ctx.font_stretch = 0xDEADBEEF
        with raises(ValueError):
            ctx.font_stretch = 'not-a-stretch-type'


def test_set_get_font_style():
    with Drawing() as ctx:
        ctx.font_style = 'italic'
        assert ctx.font_style == 'italic'
        with raises(TypeError):
            ctx.font_style = 0xDEADBEEF
        with raises(ValueError):
            ctx.font_style = 'not-a-style-type'


def test_set_get_font_weight():
    with Drawing() as ctx:
        ctx.font_weight = 400  # Normal
        assert ctx.font_weight == 400
        with raises(TypeError):
            ctx.font_weight = '400'


def test_set_get_fill_color():
    with Drawing() as ctx:
        with Color('#333333') as black:
            ctx.fill_color = black
        assert ctx.fill_color == Color('#333333')
        ctx.fill_color = 'pink'
        ctx.fill_color == Color('PINK')


def test_set_get_stroke_color():
    with Drawing() as ctx:
        with Color('#333333') as black:
            ctx.stroke_color = black
        assert ctx.stroke_color == Color('#333333')
        ctx.stroke_color = 'skyblue'
        assert ctx.stroke_color == Color('SkyBlue')


def test_set_get_stroke_width():
    with Drawing() as ctx:
        ctx.stroke_width = 5
        assert ctx.stroke_width == 5


def test_set_get_text_alignment():
    with Drawing() as ctx:
        ctx.text_alignment = 'center'
        assert ctx.text_alignment == 'center'
        with raises(TypeError):
            ctx.text_alignment = 0xDEADBEEF
        with raises(ValueError):
            ctx.text_alignment = 'not-a-text-alignment-type'


def test_set_get_text_antialias():
    with Drawing() as ctx:
        ctx.text_antialias = True
        assert ctx.text_antialias is True


def test_set_get_text_decoration():
    with Drawing() as ctx:
        ctx.text_decoration = 'underline'
        assert ctx.text_decoration == 'underline'
        with raises(TypeError):
            ctx.text_decoration = 0xDEADBEEF
        with raises(ValueError):
            ctx.text_decoration = 'not-a-text-decoration-type'


@mark.skipif(MAGICK_VERSION_NUMBER < 0x689,
             reason='DrawGetTextDirection not supported.')
def test_set_get_text_direction():
    with Drawing() as ctx:
        ctx.text_direction = 'right_to_left'
        assert ctx.text_direction == 'right_to_left'


def test_set_get_text_encoding():
    with Drawing() as ctx:
        ctx.text_encoding = 'UTF-8'
        assert ctx.text_encoding == 'UTF-8'
        ctx.text_encoding = None


def test_set_get_text_interline_spacing():
    with Drawing() as ctx:
        ctx.text_interline_spacing = 10.11
        assert ctx.text_interline_spacing == 10.11
        with raises(TypeError):
            ctx.text_interline_spacing = '10.11'


def test_set_get_text_interword_spacing():
    with Drawing() as ctx:
        ctx.text_interword_spacing = 5.55
        assert ctx.text_interword_spacing == 5.55
        with raises(TypeError):
            ctx.text_interline_spacing = '5.55'


def test_set_get_text_kerning():
    with Drawing() as ctx:
        ctx.text_kerning = 10.22
        assert ctx.text_kerning == 10.22
        with raises(TypeError):
            ctx.text_kerning = '10.22'


def test_set_get_text_under_color():
    with Drawing() as ctx:
        with Color('#333333') as black:
            ctx.text_under_color = black
        assert ctx.text_under_color == Color('#333333')
        ctx.text_under_color = '#333'  # Smoke test
        with raises(TypeError):
            ctx.text_under_color = 0xDEADBEEF


def test_set_get_vector_graphics():
    with Drawing() as ctx:
        ctx.stroke_width = 7
        xml = ctx.vector_graphics
        assert xml.index("<stroke-width>7</stroke-width>") > 0
        ctx.vector_graphics = '<wand><stroke-width>8</stroke-width></wand>'
        xml = ctx.vector_graphics
        assert xml.index("<stroke-width>8</stroke-width>") > 0
        with raises(TypeError):
            ctx.vector_graphics = 0xDEADBEEF


def test_set_get_gravity():
    with Drawing() as ctx:
        ctx.gravity = 'center'
        assert ctx.gravity == 'center'
        with raises(TypeError):
            ctx.gravity = 0xDEADBEEF
        with raises(ValueError):
            ctx.gravity = 'not-a-gravity-type'


def test_clone_drawing_wand():
    with Drawing() as ctx:
        ctx.text_kerning = 10.22
        funcs = (lambda img: Drawing(drawing=ctx),
                 lambda img: ctx.clone())
        for func in funcs:
            with func(ctx) as cloned:
                assert ctx.resource is not cloned.resource
                assert ctx.text_kerning == cloned.text_kerning


def test_clear_drawing_wand():
    with Drawing() as ctx:
        ctx.text_kerning = 10.22
        assert ctx.text_kerning == 10.22
        ctx.clear()
        assert ctx.text_kerning == 0


def test_composite():
    with Image(width=50, height=50, background='white') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.fill_color = 'black'
            ctx.stroke_color = 'black'
            ctx.rectangle(25, 25, 49, 49)
            ctx.draw(img)
            ctx.composite('replace', 0, 0, 25, 25, img)
            ctx.draw(img)
        assert was != img.signature


def test_draw_arc():
    with Image(width=50, height=50, background='white') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.fill_color = 'red'
            ctx.stroke_color = 'black'
            ctx.arc((10, 10),   # Start
                    (40, 40),   # End
                    (-90, 90))  # Degree
            ctx.draw(img)
        assert was != img.signature


def test_draw_circle():
    with Image(width=50, height=50, background='white') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.fill_color = 'black'
            ctx.circle((25, 25),  # Origin
                       (40, 40))  # Perimeter
            ctx.draw(img)
        assert was != img.signature


def test_draw_color():
    with Image(width=50, height=50, background='white') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.fill_color = 'black'
            ctx.color(25, 25, 'floodfill')
            ctx.draw(img)
        assert was != img.signature


def test_draw_color_user_error():
    with Drawing() as draw:
        with raises(TypeError):
            draw.color('apples')
        with raises(TypeError):
            draw.color(1, 2, 4)
        with raises(ValueError):
            draw.color(1, 2, 'apples')


def test_draw_ellipse():
    with Image(width=50, height=50, background='#ccc') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.fill_color = 'red'
            ctx.ellipse((25, 25),  # origin
                        (20, 10))  # radius
            ctx.draw(img)
        assert was != img.signature


def test_draw_line():
    with Image(width=10, height=10, background='#ccc') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.fill_color = 'black'
            ctx.line((5, 5), (7, 5))
            ctx.draw(img)
        assert was != img.signature


@mark.skipif(MAGICK_VERSION_NUMBER >= 0x700,
             reason='wand.drawing.Drawing.matte removed with IM 7.')
def test_draw_matte():
    white = Color('rgba(0, 255, 255, 5%)')
    transparent = Color('transparent')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_opacity = 0.0
            draw.matte(25, 25, 'floodfill')
            draw.draw(img)
            assert img[25, 25] == transparent


@mark.skipif(MAGICK_VERSION_NUMBER >= 0x700,
             reason='wand.drawing.Drawing.matte removed with IM 7.')
def test_draw_matte_user_error():
    with Drawing() as draw:
        with raises(TypeError):
            draw.matte('apples')
        with raises(TypeError):
            draw.matte(1, 2, 4)
        with raises(ValueError):
            draw.matte(1, 2, 'apples')


@mark.skipif(MAGICK_VERSION_NUMBER < 0x700,
             reason='wand.drawing.Drawing.alpha was added with IM 7.')
def test_draw_alpha():
    transparent = Color('transparent')
    with Image(width=50, height=50, pseudo='xc:white') as img:
        with Drawing() as draw:
            draw.fill_color = transparent
            draw.alpha(25, 25, 'floodfill')
            draw.draw(img)
        assert img[25, 25] == transparent


@mark.skipif(MAGICK_VERSION_NUMBER < 0x700,
             reason='wand.drawing.Drawing.alpha was added with IM 7.')
def test_draw_alpha_user_error():
    with Drawing() as draw:
        with raises(TypeError):
            draw.alpha()
        with raises(TypeError):
            draw.alpha(1, 2, 4)
        with raises(ValueError):
            draw.alpha(1, 2, 'apples')


def test_draw_point():
    white = Color('WHITE')
    black = Color('BLACK')
    with Image(width=5, height=5, background=white) as img:
        with Drawing() as draw:
            draw.stroke_color = black
            draw.point(2, 2)
            draw.draw(img)
            assert img[2, 2] == black


def test_draw_polygon():
    white = Color('WHITE')
    red = Color('RED')
    blue = Color('BLUE')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.polygon([(10, 10),
                          (40, 25),
                          (10, 40)])
            draw.draw(img)
            assert img[10, 25] == red
            assert img[25, 25] == blue
            assert img[35, 15] == img[35, 35] == white


def test_draw_polyline():
    with Image(width=50, height=50, background='white') as img:
        was = img.signature
        with Drawing() as draw:
            draw.fill_color = 'blue'
            draw.stroke_color = 'red'
            draw.polyline([(10, 10), (40, 25), (10, 40)])
            draw.draw(img)
        assert was != img.signature


def test_draw_push_pop():
    with Drawing() as draw:
        draw.stroke_width = 2
        draw.push()
        draw.stroke_width = 3
        assert 3 == draw.stroke_width
        draw.pop()
        assert 2 == draw.stroke_width


def test_draw_bezier():
    white = Color('WHITE')
    red = Color('RED')
    blue = Color('BLUE')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.bezier([(10, 10),
                         (10, 40),
                         (40, 10),
                         (40, 40)])
            draw.draw(img)
            assert img[10, 10] == img[25, 25] == img[40, 40] == red
            assert img[34, 32] == img[15, 18] == blue
            assert img[34, 38] == img[15, 12] == white


def test_path_curve():
    white = Color('WHITE')
    red = Color('RED')
    blue = Color('BLUE')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.path_start()
            draw.path_move(to=(0, 25), relative=True)
            draw.path_curve(to=(25, 25),
                            controls=((0, 0), (25, 0)))
            draw.path_curve(to=(25, 0),
                            controls=((0, 25), (25, 25)),
                            relative=True)
            draw.path_finish()
            draw.draw(img)
            assert img[25, 25] == red
            assert img[35, 35] == img[35, 35] == blue
            assert img[35, 15] == img[15, 35] == white


def test_path_curve_user_error():
    with Drawing() as draw:
        with raises(TypeError):
            draw.path_curve(to=(5, 7))
        with raises(TypeError):
            draw.path_curve(controls=(5, 7))


def test_path_curve_to_quadratic_bezier():
    white = Color('WHITE')
    red = Color('RED')
    blue = Color('BLUE')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.path_start()
            draw.path_move(to=(0, 25), relative=True)
            draw.path_curve_to_quadratic_bezier(to=(50, 25),
                                                control=(25, 50))
            draw.path_curve_to_quadratic_bezier(to=(-20, -20),
                                                control=(-25, 0),
                                                relative=True)
            draw.path_finish()
            draw.draw(img)
            assert img[30, 5] == red


def test_path_curve_to_quadratic_bezier_smooth():
    white = Color('WHITE')
    red = Color('RED')
    blue = Color('BLUE')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.path_start()
            draw.path_curve_to_quadratic_bezier(to=(25, 25),
                                                control=(25, 25))
            draw.path_curve_to_quadratic_bezier(to=(10, -10),
                                                smooth=True,
                                                relative=True)
            draw.path_curve_to_quadratic_bezier(to=(35, 35),
                                                smooth=True,
                                                relative=False)
            draw.path_curve_to_quadratic_bezier(to=(-10, -10),
                                                smooth=True,
                                                relative=True)
            draw.path_finish()
            draw.draw(img)
            assert img[25, 25] == red
            assert img[30, 30] == blue


def test_path_curve_quadratic_bezier_user_error():
    with Drawing() as draw:
        with raises(TypeError):
            draw.path_curve_to_quadratic_bezier()
        with raises(TypeError):
            draw.path_curve_to_quadratic_bezier(to=(5, 6))


def test_draw_path_elliptic_arc():
    white = Color('WHITE')
    red = Color('RED')
    blue = Color('BLUE')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.path_start()
            draw.path_move(to=(25, 0))
            draw.path_elliptic_arc(to=(25, 50), radius=(15, 25))
            draw.path_elliptic_arc(to=(0, -15), radius=(5, 5),
                                   clockwise=False, relative=True)
            draw.path_close()
            draw.path_finish()
            draw.draw(img)
            assert img[25, 35] == img[25, 20] == red
            assert img[15, 25] == img[30, 45] == blue


def test_draw_path_elliptic_arc_user_error():
    with Drawing() as draw:
        with raises(TypeError):
            draw.path_elliptic_arc(to=(5, 7))
        with raises(TypeError):
            draw.path_elliptic_arc(radius=(5, 7))


def test_draw_path_line():
    white = Color('#FFF')
    red = Color('#F00')
    blue = Color('#00F')
    with Image(width=50, height=50, background=white) as img:
        with Drawing() as draw:
            draw.fill_color = blue
            draw.stroke_color = red
            draw.stroke_width = 10
            draw.path_start()
            draw.path_move(to=(10, 10))
            draw.path_line(to=(40, 40))
            draw.path_line(to=(0, -10), relative=True)
            draw.path_horizontal_line(x=45)
            draw.path_vertical_line(y=25)
            draw.path_horizontal_line(x=-5, relative=True)
            draw.path_vertical_line(y=-5, relative=True)
            draw.path_close()
            draw.path_finish()
            draw.draw(img)
        assert img[40, 40] == img[40, 30] == red
        assert img[45, 25] == img[40, 20] == red


def test_draw_path_line_user_error():
    with Drawing() as draw:
        # Test missing value
        with raises(TypeError):
            draw.path_line()
        with raises(TypeError):
            draw.path_horizontal_line()
        with raises(TypeError):
            draw.path_vertical_line()


def test_draw_move_user_error():
    with Drawing() as draw:
        # Test missing value
        with raises(TypeError):
            draw.path_move()


@mark.parametrize('kwargs', itertools.product(
    [('right', 40), ('width', 30)],
    [('bottom', 40), ('height', 30)]
))
def test_draw_rectangle(kwargs):
    white = Color('WHITE')
    black = Color('BLACK')
    gray = Color('#ccc')
    with Image(width=50, height=50, background=white) as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.stroke_width = 2
            ctx.fill_color = black
            ctx.stroke_color = gray
            ctx.rectangle(left=10, top=10, **dict(kwargs))
            ctx.draw(img)
        assert was != img.signature


@mark.parametrize('kwargs', itertools.product(
    [('xradius', 10), ('yradius', 10)],
    [('xradius', 20)],
    [('yradius', 20)],
    [('radius', 10)]
))
def test_draw_rectangle_with_radius(kwargs):
    with Image(width=50, height=50, background='white') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.stroke_width = 2
            ctx.fill_color = 'black'
            ctx.stroke_color = '#ccc'
            ctx.rectangle(left=10, top=10,
                          width=30, height=30, **dict(kwargs))
            ctx.draw(img)
        assert was != img.signature


def test_draw_rotate():
    with Image(width=50, height=50, background='white') as img:
        was = img.signature
        with Drawing() as draw:
            draw.stroke_color = 'black'
            draw.rotate(45)
            draw.line((3, 3), (35, 35))
            draw.draw(img)
        assert was != img.signature


def test_draw_scale():
    with Image(width=50, height=50, background='white') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.fill_color = 'black'
            ctx.scale(x=2.0, y=0.5)
            ctx.rectangle(top=5, left=5, width=20, height=20)
            ctx.draw(img)
        assert was != img.signature


def test_set_fill_pattern_url():
    with Image(width=50, height=50, background='white') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.push_pattern('green_circle', 0, 0, 10, 10)
            ctx.fill_color = 'green'
            ctx.stroke_color = 'black'
            ctx.circle(origin=(5, 5), perimeter=(5, 0))
            ctx.pop_pattern()
            ctx.set_fill_pattern_url('#green_circle')
            ctx.rectangle(top=5, left=5, width=40, height=40)
            ctx.draw(img)
        assert was != img.signature


def test_set_stroke_pattern_url():
    with Image(width=50, height=50, background='white') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.push_pattern('green_ring', 0, 0, 6, 6)
            ctx.fill_color = 'green'
            ctx.stroke_color = 'white'
            ctx.circle(origin=(3, 3), perimeter=(3, 0))
            ctx.pop_pattern()
            ctx.set_stroke_pattern_url('#green_ring')
            ctx.stroke_width = 6
            ctx.rectangle(top=5, left=5, width=40, height=40)
            ctx.draw(img)
        assert was != img.signature


def test_draw_skew():
    with Image(width=50, height=50, background='white') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.stroke_color = 'black'
            ctx.skew(x=11, y=-24)
            ctx.line((3, 3), (35, 35))
            ctx.draw(img)
        assert was != img.signature


def test_draw_translate():
    with Image(width=50, height=50, background='white') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.stroke_color = 'black'
            ctx.translate(x=5, y=5)
            ctx.line((3, 3), (35, 35))
            ctx.draw(img)
        assert was != img.signature


def test_draw_text(fx_asset):
    with Image(width=100, height=100, background='white') as img:
        was = img.signature
        with Drawing() as ctx:
            ctx.font = str(fx_asset.joinpath('League_Gothic.otf'))
            ctx.font_size = 25
            ctx.fill_color = 'black'
            ctx.gravity = 'west'
            ctx.text(0, 0, 'Hello Wand')
            ctx.draw(img)
        assert was != img.signature


def test_get_font_metrics(fx_asset):
    with Image(width=144, height=192, background=Color('#fff')) as img:
        with Drawing() as draw:
            draw.font = str(fx_asset.joinpath('League_Gothic.otf'))
            draw.font_size = 13
            nm1 = draw.get_font_metrics(img, 'asdf1234')
            nm2 = draw.get_font_metrics(img, 'asdf1234asdf1234')
            nm3 = draw.get_font_metrics(img, 'asdf1234\nasdf1234')
            assert nm1.character_width == draw.font_size
            assert nm1.text_width < nm2.text_width
            assert nm2.text_width <= nm3.text_width
            assert nm2.text_height == nm3.text_height
            m1 = draw.get_font_metrics(img, 'asdf1234', True)
            m2 = draw.get_font_metrics(img, 'asdf1234asdf1234', True)
            m3 = draw.get_font_metrics(img, 'asdf1234\nasdf1234', True)
            assert m1.character_width == draw.font_size
            assert m1.text_width < m2.text_width
            assert m2.text_width > m3.text_width
            assert m2.text_height < m3.text_height


def test_viewbox():
    with Drawing() as draw:
        with raises(TypeError):
            draw.viewbox(None, None, None, None)
        with raises(TypeError):
            draw.viewbox(10, None, None, None)
        with raises(TypeError):
            draw.viewbox(10, 10, None, None)
        with raises(TypeError):
            draw.viewbox(10, 10, 100, None)
        draw.viewbox(10, 10, 100, 100)


def test_regression_issue_163(tmp_path):
    """https://github.com/emcconville/wand/issues/163"""
    unicode_char = b'\xce\xa6'.decode('utf-8')
    with Drawing() as draw:
        with Image(width=500, height=500) as image:
            draw.font_size = 20
            draw.gravity = 'south_west'
            draw.text(0, 0, unicode_char)
            draw(image)
            image.save(filename=str(tmp_path / 'out.jpg'))


def test_set_get_fill_opacity():
    with Drawing() as ctx:
        ctx.fill_opacity = 1.0
        assert ctx.fill_opacity == 1.0


def test_set_get_fill_opacity_user_error():
    with Drawing() as ctx:
        with raises(TypeError):
            ctx.fill_opacity = "1.5"


def test_set_get_fill_rule():
    with Drawing() as ctx:
        valid = 'evenodd'
        notvalid = 'error'
        invalid = (1, 2)
        ctx.fill_rule = valid
        assert ctx.fill_rule == valid
        with raises(ValueError):
            ctx.fill_rule = notvalid
        with raises(TypeError):
            ctx.fill_rule = invalid
        ctx.fill_rule = 'undefined'  # reset


@mark.skipif(MAGICK_VERSION_NUMBER < 0x700,
             reason='DrawGetOpacity always returns 1.0')
def test_set_get_opacity():
    with Drawing() as ctx:
        assert ctx.opacity == 1.0
        ctx.push()
        ctx.opacity = 0.5
        ctx.push()
        ctx.opacity = 0.25
        assert 0.24 < ctx.opacity < 0.26  # Expect float precision issues
        ctx.pop()
        assert 0.49 < ctx.opacity < 0.51  # Expect float precision issues
        ctx.pop()


def test_set_get_stroke_antialias():
    with Drawing() as ctx:
        ctx.stroke_antialias = False
        assert not ctx.stroke_antialias


def test_set_get_stroke_dash_array():
    with Drawing() as ctx:
        dash_array = [2, 1, 4, 1]
        ctx.stroke_dash_array = dash_array
        assert ctx.stroke_dash_array == dash_array


def test_set_get_stroke_dash_offset():
    with Drawing() as ctx:
        ctx.stroke_dash_offset = 0.5
        assert ctx.stroke_dash_offset == 0.5


def test_set_get_stroke_line_cap():
    with Drawing() as ctx:
        ctx.stroke_line_cap = 'round'
        assert ctx.stroke_line_cap == 'round'


def test_set_get_stroke_line_cap_user_error():
    with Drawing() as ctx:
        with raises(TypeError):
            ctx.stroke_line_cap = 0x74321870
        with raises(ValueError):
            ctx.stroke_line_cap = 'apples'


def test_set_get_stroke_line_join():
    with Drawing() as ctx:
        ctx.stroke_line_join = 'miter'
        assert ctx.stroke_line_join == 'miter'


def test_set_get_stroke_line_join_user_error():
    with Drawing() as ctx:
        with raises(TypeError):
            ctx.stroke_line_join = 0x74321870
        with raises(ValueError):
            ctx.stroke_line_join = 'apples'


def test_set_get_stroke_miter_limit():
    with Drawing() as ctx:
        ctx.stroke_miter_limit = 5
        assert ctx.stroke_miter_limit == 5


def test_set_get_stroke_miter_limit_user_error():
    with Drawing() as ctx:
        with raises(TypeError):
            ctx.stroke_miter_limit = '5'


def test_set_get_stroke_opacity():
    with Drawing() as ctx:
        ctx.stroke_opacity = 1.0
        assert ctx.stroke_opacity == 1.0


def test_set_get_stroke_opacity_user_error():
    with Drawing() as ctx:
        with raises(TypeError):
            ctx.stroke_opacity = '1.0'


def test_set_get_stroke_width_user_error():
    with Drawing() as ctx:
        with raises(TypeError):
            ctx.stroke_width = '0.1234'
        with raises(ValueError):
            ctx.stroke_width = -1.5


def test_draw_affine():
    with Image(width=100, height=100, background='skyblue') as img:
        was = img.signature
        img.format = 'png'
        with Drawing() as ctx:
            ctx.affine([1.5, 0.5, 0, 1.5, 45, 25])
            ctx.rectangle(top=5, left=5, width=25, height=25)
            ctx.draw(img)
        assert was != img.signature
    with raises(ValueError):
        with Drawing() as ctx:
            ctx.affine([1.0])
    with raises(TypeError):
        with Drawing() as ctx:
            ctx.affine(['a', 'b', 'c', 'd', 'e', 'f'])


def test_draw_clip_path():
    skyblue = Color('skyblue')
    orange = Color('orange')
    with Image(width=100, height=100, background='skyblue') as img:
        with Drawing() as ctx:
            ctx.push_defs()
            ctx.push_clip_path("eyes_only")
            ctx.push()
            ctx.rectangle(top=0, left=0, width=50, height=50)
            ctx.pop()
            ctx.pop_clip_path()
            ctx.pop_defs()
            ctx.clip_path = "eyes_only"
            ctx.clip_rule = "nonzero"
            ctx.clip_path_units = "object_bounding_box"
            ctx.fill_color = orange
            ctx.rectangle(top=5, left=5, width=90, height=90)
            ctx.draw(img)
        assert img[75, 75] == skyblue
